# -*- coding: utf-8 -*-
# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

# Sadly have to import the whole "nemo" python module to avoid circular dependencies.
# Moreover, at that point nemo module doesn't contain "core", so during "python module registration"
# nothing from nemo.core, including e.g. types (so we cannot use them for "python 3 type hints").
import nemo
from nemo.utils.metaclasses import Singleton
from nemo.utils.neural_graph.neural_graph_manager import NeuralGraphManager
from nemo.utils.neural_graph.object_registry import ObjectRegistry
from nemo.utils.nmtensor_registry import NmTensorNameRegistry


class AppState(metaclass=Singleton):
    """
        Application state stores variables important from the point of view of execution of the NeMo application.
        Staring from the most elementary (epoch number, episode number, device used etc.) to the currently
        active graph etc.
    """

    def __init__(self, device=None):
        """
            Constructor. Initializes global variables.

            Args:
                device: main device used for computations [CPU | GPU] (DEFAULT: GPU)
        """
        # Had to set it to None in argument to avoid circular import at the class initialization phase.
        if device is None:
            self._device = nemo.core.DeviceType.GPU
        else:
            self._device = device
        # Create module registry.
        self._module_registry = ObjectRegistry("module")
        # Create graph manager (registry with some additional functionality).
        self._neural_graph_manager = NeuralGraphManager()
        # Create NmTensor registry
        self._nmtensor_name_registry = NmTensorNameRegistry()

    @property
    def tensor_names(self):
        """ Property returning the NmTensorNameRegistry which maps user-defined names to tensor's unique_names.

            Returns:
                NmTensorNameRegistry.
        """
        return self._nmtensor_name_registry

    @property
    def modules(self):
        """
            Property returning the existing modules.

            Returns:
                Existing modules (a set object).
        """
        return self._module_registry

    @property
    def graphs(self):
        """ Property returning the existing graphs.

            Returns:
                Existing graphs (a set object).
        """
        return self._neural_graph_manager

    def register_module(self, module, name: str) -> str:
        """
            Registers a module using the provided name.
            If name is none - generates a new unique name.

            Args:
                module: A Neural Module object to be registered.
                name: A "proposition" of module name.

            Returns:
                A unique name (proposition or newly generated name).
        """
        return self._module_registry.register(module, name)

    def register_graph(self, graph, name: str) -> str:
        """
            Registers a new graph using the provided name.
            If name is none - generates a new unique name.

            Args:
                graph: A Neural Graph object to be registered.
                name: A "proposition" of graph name.

            Returns:
                A unique name (proposition or newly generated name).
        """
        return self._neural_graph_manager.register(graph, name)

    @property
    def active_graph(self):
        """ Property returns the active graph.

            Returns:
                Active graph.
        """
        return self._neural_graph_manager.active_graph

    @active_graph.setter
    def active_graph(self, graph):
        """ Property sets the active graph.

            Args:
                graph: Neural graph object that will become active.
        """
        self._neural_graph_manager.active_graph = graph



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

'''
This file contains code artifacts adapted from the original implementation:
https://github.com/thu-coai/ConvLab-2/
'''

from nemo.backends.pytorch.nm import NonTrainableNM
from nemo.collections.nlp.neural_types import *
from nemo.core.neural_types import *
from nemo.utils import logging
from nemo.utils.decorators import add_port_docs

__all__ = ['SystemUtteranceHistoryUpdate']


class SystemUtteranceHistoryUpdate(NonTrainableNM):
    """
    Updates dialogue history with system utterance.
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        user_uttr (str): user utterance
        sys_uttr (str): system utterace
        dialog_history (list): dialogue history, list of system and diaglogue utterances
        """
        return {
            'sys_uttr': NeuralType(
                axes=[AxisType(kind=AxisKind.Batch, is_list=True)], elements_type=SystemUtterance()
            ),
            'dialog_history': NeuralType(
                axes=(AxisType(kind=AxisKind.Batch, is_list=True), AxisType(kind=AxisKind.Time, is_list=True),),
                elements_type=AgentUtterance(),
            ),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        dialog_history (list): dialogue history, being a list of user and system utterances.
        """
        return {
            'dialog_history': NeuralType(
                axes=(AxisType(kind=AxisKind.Batch, is_list=True), AxisType(kind=AxisKind.Time, is_list=True),),
                elements_type=AgentUtterance(),
            ),
        }

    def __init__(self):
        """
        Initializes the object
        Args:
            data_desc (obj): data descriptor for MultiWOZ dataset, contains information about domains, slots, 
                    and associated vocabulary
        """
        super().__init__()

    def forward(self, sys_uttr, dialog_history):
        """
        Returns updated dialog history.
        Args:
            sys_uttr (str): system utterace
            dialog_history (list): dialogue history, list of user and system diaglogue utterances
        Returns:
            dialog_history (list): updated dialogue history, list of user and system diaglogue utterances
        """
        dialog_history.append(["sys", sys_uttr])
        logging.debug("Dialogue history: %s", dialog_history)

        return dialog_history



# -*- coding: utf-8 -*-
# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
# Copyright (C) tkornuta, IBM Corporation 2019
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

__author__ = "Tomasz Kornuta"

"""
This file contains code artifacts adapted from the original implementation:
https://github.com/IBM/pytorchpipe/blob/develop/ptp/components/models/general_usage/feed_forward_network.py
"""

from typing import List, Optional

import torch

from nemo.backends.pytorch.nm import TrainableNM
from nemo.core.neural_types import AxisKind, AxisType, NeuralType, VoidType
from nemo.utils import logging
from nemo.utils.configuration_error import ConfigurationError
from nemo.utils.decorators import add_port_docs

__all__ = ['FeedForwardNetwork']


class FeedForwardNetwork(TrainableNM):
    """
    A simple trainable module consisting of several stacked fully connected layers
    with ReLU non-linearities and dropout between them.
    
    # TODO: parametrize with other non-linearities.
    
    """

    def __init__(
        self,
        input_size: int,
        output_size: int,
        hidden_sizes: List[int] = [],
        dimensions: int = 2,
        dropout_rate: float = 0,
        name: Optional[str] = None,
    ):
        """
        Initializes the feed-forwad network.

        Args:
            input_size: Size of input (1D)
            output_sizes: Size of the output (1D)
            hidden_sizes: Sizes of the consecutive hidden layers (DEFAULT: [] = no hidden)
            dimensions: Number of dimensions of input/output tensors (DEFAULT: 2 = BATCH X INPUT_SIZE)
            dropout_rate: Dropout rage (Default: 0)
            name: Name of the module (DEFAULT: None)
        """
        # Call constructor of the parent class.
        TrainableNM.__init__(self, name=name)

        # Get input size.
        self._input_size = input_size
        if type(self._input_size) == list:
            if len(self._input_size) == 1:
                self._input_size = self._input_size[0]
            else:
                raise ConfigurationError("'input_size' must be a single value (received {})".format(self._input_size))

        # Get input/output dimensions, i.e. number of axes of the input [BATCH_SIZE x ... x INPUT_SIZE].
        # The module will "broadcast" over those dimensions.
        self._dimensions = dimensions
        if self._dimensions < 2:
            raise ConfigurationError("'dimensions' must be bigger than two  (received {})".format(self._dimensions))

        # Get output (prediction/logits) size.
        self._output_size = output_size
        if type(self._output_size) == list:
            if len(self._output_size) == 1:
                self._output_size = self._output_size[0]
            else:
                raise ConfigurationError(
                    "'output_size' must be a single value (received {})".format(self._output_size)
                )

        logging.info(
            "Initializing network with input size = {} and output size = {}".format(
                self._input_size, self._output_size
            )
        )

        # Create the module list.
        modules = []

        # Retrieve number of hidden layers, along with their sizes (numbers of hidden neurons from configuration).
        if type(hidden_sizes) == list:
            # Stack linear layers.
            input_dim = self._input_size
            for hidden_dim in hidden_sizes:
                # Add linear layer.
                modules.append(torch.nn.Linear(input_dim, hidden_dim))
                # Add activation.
                modules.append(torch.nn.ReLU())
                # Add dropout.
                if dropout_rate > 0:
                    modules.append(torch.nn.Dropout(dropout_rate))
                # Remember size.
                input_dim = hidden_dim

            # Add the last output" (or in a special case: the only) layer.
            modules.append(torch.nn.Linear(input_dim, self._output_size))

            logging.info("Created {} hidden layers with sizes {}".format(len(hidden_sizes), hidden_sizes))

        else:
            raise ConfigurationError(
                "'hidden_sizes' must contain a list with numbers of neurons in consecutive hidden layers (received {})".format(
                    hidden_sizes
                )
            )

        # Finally create the sequential model out of those modules.
        self.layers = torch.nn.Sequential(*modules)

    @property
    @add_port_docs()
    def input_ports(self):
        """
        Returns definitions of module input ports.
        Batch of inputs, each represented as index [BATCH_SIZE x ... x INPUT_SIZE]
        """
        # Prepare list of axes.
        axes = [AxisType(kind=AxisKind.Batch)]
        # Add the "additional dimensions".
        for _ in range(self._dimensions)[1:-1]:
            axes.append(AxisType(kind=AxisKind.Any))
        # Add the last axis: input_size
        axes.append(AxisType(kind=AxisKind.Any, size=self._input_size))
        # Return neural type.
        return {"inputs": NeuralType(axes, VoidType())}

    @property
    @add_port_docs()
    def output_ports(self):
        """
        Returns definitions of module output ports.
        """
        # Prepare list of axes.
        axes = [AxisType(kind=AxisKind.Batch)]
        # Add the "additional dimensions".
        for _ in range(self._dimensions)[1:-1]:
            axes.append(AxisType(kind=AxisKind.Any))
        # Add the last axis: input_size
        axes.append(AxisType(kind=AxisKind.Any, size=self._output_size))
        # Return neural type: batch of "logits" of "any type".
        return {"outputs": NeuralType(axes, VoidType())}

    def forward(self, inputs):
        """
        Performs the forward step of the module.

        Args:
            inputs: Batch of inputs to be processed [BATCH_SIZE x ... x INPUT_SIZE]
        
        Returns:
            Batch of outputs/predictions (log_probs) [BATCH_SIZE x ... x NUM_CLASSES]
        """

        # print("{}: input shape: {}, device: {}\n".format(self.name, inputs.shape, inputs.device))

        # Check that the input has the number of dimensions that we expect
        if len(inputs.shape) != self._dimensions:
            raise ConfigurationError(
                "Expected `{}` dimensions for input, but received `{}` instead. "
                F"Check fix the dimensions in your script.".format(self._dimensions, len(inputs.shape))
            )

        # Reshape such that we do a broadcast over the last dimension
        origin_shape = inputs.shape
        inputs = inputs.contiguous().view(-1, origin_shape[-1])

        # Propagate inputs through all layers and activations.
        outputs = self.layers(inputs)

        # Restore the input dimensions but the last one (as it's been resized by the FFN)
        outputs = outputs.view(*origin_shape[0 : self._dimensions - 1], -1)

        # Return the result.
        return outputs



# Copyright (c) 2019 NVIDIA Corporation
import torch as t
import torch.nn as nn

from ...core import DeviceType, NeuralModule
from ...utils.helpers import rgetattr, rsetattr


class TrainableNeuralModuleWrapper(NeuralModule, nn.Module):
    """This class wraps an instance of Pytorch's nn.Module and
    returns NeuralModule's instance."""

    def __init__(self, pt_nn_module, input_ports_dict, output_ports_dict):
        NeuralModule.__init__(self)
        nn.Module.__init__(self)
        self._input_ports = input_ports_dict
        self._output_ports = output_ports_dict
        self._device = t.device("cuda" if self.placement in [DeviceType.GPU, DeviceType.AllGpu] else "cpu")
        self._pt_module = pt_nn_module
        self._pt_module.to(self._device)

    @property
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return self._input_ports

    @property
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return self._output_ports

    # def forward(self, *input):
    #  return self._pt_module(input)

    def eval(self):
        return self._pt_module.eval()

    def train(self):
        return self._pt_module.train()

    def __call__(self, force_pt=False, *input, **kwargs):
        pt_call = len(input) > 0 or force_pt
        if pt_call:
            return self._pt_module.__call__(*input, **kwargs)
        else:
            return NeuralModule.__call__(self, **kwargs)

    def save_to(self, path):
        t.save(self._pt_module.state_dict(), path)

    def restore_from(self, path):
        self._pt_module.load_state_dict(t.load(path))

    def parameters(self):
        return self._pt_module.parameters()

    def named_parameters(self):
        return self._pt_module.named_parameters()

    def freeze(self, weights=None):
        for name, param in self._pt_module.named_parameters():
            if weights is None or name in weights:
                param.requires_grad = False

    def unfreeze(self, weights=None):
        for name, param in self._pt_module.named_parameters():
            if weights is None or name in weights:
                param.requires_grad = True

    def get_weights(self):
        result = {}
        for name, parameter in self._pt_module.named_parameters():
            result[name] = (parameter, parameter.requires_grad)
        return result

    def set_weights(self, name2weight, name2name_and_transform=None):
        if name2name_and_transform:
            raise NotImplementedError("Transforms are not currently supported for set_weights")
        self._pt_module.load_state_dict({key: name2weight[key][0] for key in name2weight.keys()})

    def tie_weights_with(self, module, weight_names):
        for name in weight_names:
            rsetattr(self._pt_module, name, nn.Parameter(rgetattr(module, name)))

    @property
    def num_weights(self):
        return sum(p.numel() for p in self._pt_module.parameters() if p.requires_grad)



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from torch import nn as nn

from nemo.backends.pytorch import MultiLayerPerceptron, TrainableNM
from nemo.collections.nlp.utils.transformer_utils import transformer_weights_init
from nemo.core import ChannelType, LogitsType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['JointIntentSlotClassifier']


class JointIntentSlotClassifier(TrainableNM):
    """
    The softmax classifier for the joint intent classification and slot
    filling task which  consists of a dense layer + relu + softmax for
    predicting the slots and similar for predicting the intents.

    Args:
        hidden_size (int): the size of the hidden state for the dense layer
        num_intents (int): number of intents
        num_slots (int): number of slots
        dropout (float): dropout to be applied to the layer
        use_transformer_pretrained (bool):
            TODO
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.

        hidden_states:
            TODO
        """
        return {"hidden_states": NeuralType(('B', 'T', 'C'), ChannelType())}

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        intent_logits:
            TODO
        slot_logits:
            TODO
        """
        return {
            "intent_logits": NeuralType(('B', 'D'), LogitsType()),
            "slot_logits": NeuralType(('B', 'T', 'D'), LogitsType()),
        }

    def __init__(self, hidden_size, num_intents, num_slots, dropout=0.0, use_transformer_pretrained=True, **kwargs):
        super().__init__(**kwargs)
        self.dropout = nn.Dropout(dropout)
        self.slot_mlp = MultiLayerPerceptron(
            hidden_size, num_classes=num_slots, device=self._device, num_layers=2, activation='relu', log_softmax=False
        )
        self.intent_mlp = MultiLayerPerceptron(
            hidden_size,
            num_classes=num_intents,
            device=self._device,
            num_layers=2,
            activation='relu',
            log_softmax=False,
        )
        if use_transformer_pretrained:
            self.apply(lambda module: transformer_weights_init(module, xavier=False))

    def forward(self, hidden_states):
        hidden_states = self.dropout(hidden_states)
        intent_logits = self.intent_mlp(hidden_states[:, 0])
        slot_logits = self.slot_mlp(hidden_states)
        return intent_logits, slot_logits



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2018 The Google AI Language Team Authors and
# The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

# Some code of this file was adapted from the HuggingFace library available at
# https://github.com/huggingface/transformers

import os
import pickle

import numpy as np
import torch
from torch.utils.data import Dataset

from nemo import logging
from nemo.collections.nlp.data.datasets.glue_benchmark_dataset.data_processors import *

__all__ = ['GLUEDataset', 'output_modes', 'processors']

processors = {
    "cola": ColaProcessor,
    "mnli": MnliProcessor,
    "mnli-mm": MnliMismatchedProcessor,
    "mrpc": MrpcProcessor,
    "sst-2": Sst2Processor,
    "sts-b": StsbProcessor,
    "qqp": QqpProcessor,
    "qnli": QnliProcessor,
    "rte": RteProcessor,
    "wnli": WnliProcessor,
}
output_modes = {
    "cola": "classification",
    "mnli": "classification",
    "mnli-mm": "classification",
    "mrpc": "classification",
    "sst-2": "classification",
    "sts-b": "regression",
    "qqp": "classification",
    "qnli": "classification",
    "rte": "classification",
    "wnli": "classification",
}
GLUE_TASKS_NUM_LABELS = {
    "cola": 2,
    "mnli": 3,
    "mrpc": 2,
    "sst-2": 2,
    "sts-b": 1,
    "qqp": 2,
    "qnli": 2,
    "rte": 2,
    "wnli": 2,
}


class GLUEDataset(Dataset):
    def __init__(self, data_dir, tokenizer, max_seq_length, processor, output_mode, evaluate, use_data_cache):

        self.tokenizer = tokenizer
        self.label_list = processor.get_labels()
        self.examples = processor.get_dev_examples(data_dir) if evaluate else processor.get_train_examples(data_dir)
        processor_name = type(processor).__name__
        tokenizer_type = type(tokenizer.tokenizer).__name__
        vocab_size = getattr(tokenizer, "vocab_size", 0)
        cached_features_file = os.path.join(
            data_dir,
            "cached_{}_{}_{}_{}_{}".format(
                processor_name, "dev" if evaluate else "train", tokenizer_type, str(max_seq_length), str(vocab_size)
            ),
        )

        if use_data_cache and os.path.exists(cached_features_file):
            logging.info(f"loading from {cached_features_file}")
            with open(cached_features_file, "rb") as reader:
                self.features = pickle.load(reader)
        else:
            token_params = {
                'bos_token': None,
                'eos_token': tokenizer.eos_token,
                'pad_token': tokenizer.pad_token,
                'cls_token': tokenizer.cls_token,
                'sep_token_extra': tokenizer.eos_token if 'roberta' in tokenizer_type.lower() else None,
            }

            self.features = self.convert_examples_to_features(
                self.examples, self.label_list, max_seq_length, tokenizer, output_mode, **token_params
            )
            if use_data_cache:
                master_device = not torch.distributed.is_initialized() or torch.distributed.get_rank() == 0
                if master_device:
                    logging.info(f'Saving train features into {cached_features_file}')
                    with open(cached_features_file, "wb") as writer:
                        pickle.dump(self.features, writer)

    def __len__(self):
        return len(self.features)

    def __getitem__(self, idx):
        feature = self.features[idx]
        return (
            np.array(feature.input_ids),
            np.array(feature.segment_ids),
            np.array(feature.input_mask, dtype=np.long),
            np.array(feature.label_id),
        )

    def convert_examples_to_features(
        self,
        examples,
        label_list,
        max_seq_length,
        tokenizer,
        output_mode,
        bos_token=None,
        eos_token='[SEP]',
        pad_token='[PAD]',
        cls_token='[CLS]',
        sep_token_extra=None,
        cls_token_at_end=False,
        cls_token_segment_id=0,
        pad_token_segment_id=0,
        pad_on_left=False,
        mask_padding_with_zero=True,
        sequence_a_segment_id=0,
        sequence_b_segment_id=1,
    ):
        """
        Loads a data file into a list of `InputBatch`s.
        The `cls_token_at_end` defines the location of the CLS token:

            * False (Default, BERT/XLM pattern): [CLS] + A + [SEP] + B + [SEP]
            * True (XLNet/GPT pattern): A + [SEP] + B + [SEP] + [CLS]

        The `cls_token_segment_id` defines the segment id associated to the CLS token (0 for BERT, 2 for XLNet)
        
        The convention in BERT is:
        
            a. For sequence pairs:
                * tokens:   [CLS] is this jack ##ville ? [SEP] no it is not . [SEP]
                * type_ids:   0   0  0    0    0       0   0   1  1  1  1   1   1
            b. For single sequences:
                * tokens:   [CLS] the dog is hairy . [SEP]
                * type_ids:   0   0   0   0  0     0   0

        Where "type_ids" are used to indicate whether this is the first
        sequence or the second sequence. The embedding vectors for `type=0`
        and `type=1` were learned during pre-training and are added to the
        wordpiece embedding vector (and position vector). This is
        not *strictly* necessarysince the [SEP] token unambiguously separates
        the sequences, but it makes it easier for the model to learn
        the concept of sequences.
        For classification tasks, the first vector (corresponding to [CLS])
        is used as as the "sentence vector". Note that this only makes sense
        because the entire model is fine-tuned.
        
        The convention for NMT is:
        
            a. For sequence pairs:
                * tokens:<BOS> is this jack ##ville ? <EOS> <BOS> no it is not . <EOS>
                * type_ids:0   0  0    0    0       0   0     1   1  1  1  1   1   1
            b. For single sequences:
                * tokens:   <BOS> the dog is hairy . <EOS>
                * type_ids:   0   0   0   0  0     0   0

        """
        label_map = {label: i for i, label in enumerate(label_list)}

        features = []
        for ex_index, example in enumerate(examples):
            if ex_index % 10000 == 0:
                logging.info("Writing example %d of %d" % (ex_index, len(examples)))

            tokens_a = tokenizer.text_to_tokens(example.text_a)

            tokens_b = None
            if example.text_b:
                tokens_b = tokenizer.text_to_tokens(example.text_b)

                special_tokens_count = 2 if eos_token else 0
                special_tokens_count += 1 if sep_token_extra else 0
                special_tokens_count += 2 if bos_token else 0
                special_tokens_count += 1 if cls_token else 0
                self._truncate_seq_pair(tokens_a, tokens_b, max_seq_length - special_tokens_count)
            else:
                special_tokens_count = 1 if eos_token else 0
                special_tokens_count += 1 if sep_token_extra else 0
                special_tokens_count += 1 if bos_token else 0
                if len(tokens_a) > max_seq_length - special_tokens_count:
                    tokens_a = tokens_a[: max_seq_length - special_tokens_count]
            # Add special tokens to sequence_a
            tokens = tokens_a
            if bos_token:
                tokens = [bos_token] + tokens
            if eos_token:
                tokens += [eos_token]
            segment_ids = [sequence_a_segment_id] * len(tokens)

            # Add sequence separator between sequences
            if tokens_b and sep_token_extra:
                tokens += [sep_token_extra]
                segment_ids += [sequence_a_segment_id]

            # Add special tokens to sequence_b
            if tokens_b:
                if bos_token:
                    tokens += [bos_token]
                    segment_ids += [sequence_b_segment_id]
                tokens += tokens_b
                segment_ids += [sequence_b_segment_id] * (len(tokens_b))
                if eos_token:
                    tokens += [eos_token]
                    segment_ids += [sequence_b_segment_id]

            # Add classification token - for BERT models
            if cls_token:
                if cls_token_at_end:
                    tokens += [cls_token]
                    segment_ids += [cls_token_segment_id]
                else:
                    tokens = [cls_token] + tokens
                    segment_ids = [cls_token_segment_id] + segment_ids
            input_ids = tokenizer.tokens_to_ids(tokens)

            # The mask has 1 for real tokens and 0 for padding tokens. Only real
            # tokens are attended to.
            input_mask = [1 if mask_padding_with_zero else 0] * len(input_ids)

            # Zero-pad up to the sequence length.
            padding_length = max_seq_length - len(input_ids)
            pad_token_id = tokenizer.tokens_to_ids([pad_token])[0]
            if pad_on_left:
                input_ids = ([pad_token_id] * padding_length) + input_ids
                input_mask = ([0 if mask_padding_with_zero else 1] * padding_length) + input_mask
                segment_ids = ([pad_token_segment_id] * padding_length) + segment_ids
            else:
                input_ids = input_ids + ([pad_token_id] * padding_length)
                input_mask = input_mask + ([0 if mask_padding_with_zero else 1] * padding_length)
                segment_ids = segment_ids + ([pad_token_segment_id] * padding_length)
            if len(input_ids) != max_seq_length:
                raise ValueError("input_ids must be of length max_seq_length")
            if len(input_mask) != max_seq_length:
                raise ValueError("input_mask must be of length max_seq_length")
            if len(segment_ids) != max_seq_length:
                raise ValueError("segment_ids must be of length max_seq_length")
            if output_mode == "classification":
                label_id = label_map[example.label]
            elif output_mode == "regression":
                label_id = np.float32(example.label)
            else:
                raise KeyError(output_mode)

            if ex_index < 5:
                logging.info("*** Example ***")
                logging.info("guid: %s" % (example.guid))
                logging.info("tokens: %s" % " ".join(list(map(str, tokens))))
                logging.info("input_ids: %s" % " ".join(list(map(str, input_ids))))
                logging.info("input_mask: %s" % " ".join(list(map(str, input_mask))))
                logging.info("segment_ids: %s" % " ".join(list(map(str, segment_ids))))
                logging.info("label: %s (id = %d)" % (example.label, label_id))

            features.append(
                InputFeatures(input_ids=input_ids, input_mask=input_mask, segment_ids=segment_ids, label_id=label_id)
            )
        return features

    def _truncate_seq_pair(self, tokens_a, tokens_b, max_length):
        """Truncates a sequence pair in place to the maximum length.

        This will always truncate the longer sequence one token at a time.
        This makes more sense than truncating an equal percent
        of tokens from each, since if one sequence is very short then each token
        that's truncated likely contains more information than a longer sequence.
        """
        while True:
            total_length = len(tokens_a) + len(tokens_b)
            if total_length <= max_length:
                break
            if len(tokens_a) > len(tokens_b):
                tokens_a.pop()
            else:
                tokens_b.pop()


class InputFeatures(object):
    """A single set of features of data."""

    def __init__(self, input_ids, input_mask, segment_ids, label_id):
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids
        self.label_id = label_id



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2019 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""Evaluation metrics for Schema-guided dialogue.

This library provides functions for calculating the evaluation metrics for a
single dialogue. The following metrics are defined:

(1) Active intent accuracy: The fraction of user turns for which the active
  intent has been correctly predicted.
(2) Slot tagging F1: The macro-averaged F1 score for tagging slot values for
  non-categorical slots. This metric is optional to report in the final paper
  if participants decide not to use slot tagging.
(3) Requested slots F1: The macro-averaged F1 score for requested slots over the
  turns. For a turn, if there are no requested slots in both the ground truth
  and the prediction, that turn is skipped. The reported number is the average
  F1 score for all un-skipped user turns. This metric is optional to report in
  the final paper.
(4) Average goal accuracy: For each turn, participants must predict a single
  value for each slot present in the dialogue state. The slots which have a
  non-empty assignment in the ground truth dialogue state are only considered.
  This is the average accuracy of predicting the value of a slot correctly. A
  fuzzy matching based score is used for non-categorical slots.
(5) Joint goal accuracy: This is the average accuracy of predicting all slot
  assignments for a turn correctly. A fuzzy matching based score is used for
  non-categorical slots. This is the primary evaluation metric used for ranking
  submissions. More details to follow with the evaluation script.

This file contains code artifacts adapted from the original implementation:
https://github.com/google-research/google-research/blob/master/schema_guided_dst/metrics.py
"""

import collections

import numpy as np
from rapidfuzz import fuzz

F1Scores = collections.namedtuple("F1Scores", ["f1", "precision", "recall"])

# Evaluation and other relevant metrics for DSTC8/SGD Schema-guided DST.
# (1) Active intent accuracy.
ACTIVE_INTENT_ACCURACY = "active_intent_accuracy"
# (2) Slot tagging F1.
SLOT_TAGGING_F1 = "slot_tagging_f1"
SLOT_TAGGING_PRECISION = "slot_tagging_precision"
SLOT_TAGGING_RECALL = "slot_tagging_recall"
# (3) Requested slots F1.
REQUESTED_SLOTS_F1 = "requested_slots_f1"
REQUESTED_SLOTS_PRECISION = "requested_slots_precision"
REQUESTED_SLOTS_RECALL = "requested_slots_recall"
# (4) Average goal accuracy.
AVERAGE_GOAL_ACCURACY = "average_goal_accuracy"
AVERAGE_CAT_ACCURACY = "average_cat_accuracy"
AVERAGE_NONCAT_ACCURACY = "average_noncat_accuracy"
# (5) Joint goal accuracy.
JOINT_GOAL_ACCURACY = "joint_goal_accuracy"
JOINT_CAT_ACCURACY = "joint_cat_accuracy"
JOINT_NONCAT_ACCURACY = "joint_noncat_accuracy"

NAN_VAL = "NA"


def compute_f1(list_ref, list_hyp):
    """Compute F1 score from reference (grouth truth) list and hypothesis list.

  Args:
    list_ref: List of true elements.
    list_hyp: List of postive (retrieved) elements.

  Returns:
    A F1Scores object containing F1, precision, and recall scores.
  """

    ref = collections.Counter(list_ref)
    hyp = collections.Counter(list_hyp)
    true = sum(ref.values())
    positive = sum(hyp.values())
    true_positive = sum((ref & hyp).values())
    precision = float(true_positive) / positive if positive else 1.0
    recall = float(true_positive) / true if true else 1.0
    if precision + recall > 0.0:
        f1 = 2.0 * precision * recall / (precision + recall)
    else:  # The F1-score is defined to be 0 if both precision and recall are 0.
        f1 = 0.0

    return F1Scores(f1=f1, precision=precision, recall=recall)


def fuzzy_string_match(str_ref, str_hyp):
    """Returns fuzzy string similarity score in range [0.0, 1.0]."""

    # The higher the score, the higher the similarity between the two strings.
    return fuzz.token_sort_ratio(str_ref, str_hyp) / 100.0


def noncat_slot_value_match(str_ref_list, str_hyp, no_fuzzy_match):
    """Calculate non-categorical slots correctness.

  Args:
    str_ref_list: a list of reference strings.
    str_hyp: the hypothesis string.
    use_fuzzy_match: whether to use fuzzy string matching.

  Returns:
    score: The highest fuzzy string match score of the references and hypotheis.
  """
    score = 0.0
    for str_ref in str_ref_list:
        if no_fuzzy_match:
            match_score = float(str_ref == str_hyp)
        else:
            match_score = fuzzy_string_match(str_ref, str_hyp)
        score = max(score, match_score)
    return score


def compare_slot_values(slot_values_ref, slot_values_hyp, service, no_fuzzy_match):
    """Compare and get correctness of goal state's slot_values.

  Args:
    slot_values_ref: goal state slot_values from reference (ground truth).
    slot_values_hyp: goal state slot_values from hypothesis (prediction).
    service: a service data structure in the schema. We use it to obtain the
      list of slots in the service and infer whether a slot is categorical.
    use_fuzzy_match: whether to use fuzzy string matching for non-categorical
      slot values

  Returns:
    (list_cor, slot_active, slot_cat)
    list_cor: list of corectness scores, each corresponding to one slot in the
        service. The score is a float either 0.0 or 1.0 for categorical slot,
        and in range [0.0, 1.0] for non-categorical slot.
    slot_active: list indicating whether the element in list_cor corresponds to
        an active ground-truth slot.
    slot_cat: list indicating whether the element in list_cor corresponds to a
        categorical slot.
  """
    list_cor = []
    slot_active = []
    slot_cat = []

    for slot in service["slots"]:
        slot_name = slot["name"]
        slot_cat.append(slot["is_categorical"])

        if slot_name in slot_values_ref:  # REF=active
            slot_active.append(True)
            if slot_name in slot_values_hyp:  # HYP=active, apply matching
                value_ref_list = slot_values_ref[slot_name]
                value_hyp = slot_values_hyp[slot_name][0]
                if slot["is_categorical"]:
                    cor = float(value_ref_list[0] == value_hyp)
                else:
                    cor = noncat_slot_value_match(value_ref_list, value_hyp, no_fuzzy_match)

                list_cor.append(cor)
            else:  # HYP=off
                list_cor.append(0.0)
        else:  # REF=off
            slot_active.append(False)
            if slot_name in slot_values_hyp:  # HYP=active
                list_cor.append(0.0)
            else:  # HYP=off
                list_cor.append(1.0)

    assert len(list_cor) == len(service["slots"])
    assert len(slot_active) == len(service["slots"])
    assert len(slot_cat) == len(service["slots"])
    return list_cor, slot_active, slot_cat


def get_active_intent_accuracy(frame_ref, frame_hyp):
    """Get active intent accuracy of a frame.

  Args:
    frame_ref: single semantic frame from reference (ground truth) file.
    frame_hyp: single semantic frame from hypothesis (prediction) file.

  Returns:
    1.0 if the intent prediction is correct, otherwise 0.0.
  """
    return float(frame_ref["state"]["active_intent"] == frame_hyp["state"]["active_intent"])


def get_slot_tagging_f1(frame_ref, frame_hyp, utt, service):
    """Get slot tagging (non-categorical slots only) F1 scores of a frame.

  Args:
    frame_ref: single semantic frame from reference (ground truth) file.
    frame_hyp: single semantic frame from hypothesis (prediction) file.
    utt: user utterance. Slot tagging annotations are the character positions in
      the utterance.
    service: a service data structure in the schema. We use it to infer whether
      a slot is non-categorical.

  Returns:
    A F1Scores object containing F1, precision, and recall scores.
  """

    list_noncat_slots = [s["name"] for s in service["slots"] if not s["is_categorical"]]
    if "slots" not in frame_hyp:
        return None
    else:
        list_ref = [
            (s["slot"], utt[s["start"] : s["exclusive_end"]])
            for s in frame_ref["slots"]
            if s["slot"] in list_noncat_slots
        ]
        list_hyp = [
            (s["slot"], utt[s["start"] : s["exclusive_end"]])
            for s in frame_hyp["slots"]
            if s["slot"] in list_noncat_slots
        ]
        return compute_f1(list_ref, list_hyp)


def get_requested_slots_f1(frame_ref, frame_hyp):
    """Get requested slots F1 scores of a frame.

  Args:
    frame_ref: single semantic frame from reference (ground truth) file.
    frame_hyp: single semantic frame from hypothesis (prediction) file.

  Returns:
    A F1Scores object containing F1, precision, and recall scores.
  """
    return compute_f1(frame_ref["state"]["requested_slots"], frame_hyp["state"]["requested_slots"])


def get_average_and_joint_goal_accuracy(frame_ref, frame_hyp, service, no_fuzzy_match):
    """Get average and joint goal accuracies of a frame.

  Args:
    frame_ref: single semantic frame from reference (ground truth) file.
    frame_hyp: single semantic frame from hypothesis (prediction) file.
    service: a service data structure in the schema. We use it to obtain the
      list of slots in the service and infer whether a slot is categorical.
    use_fuzzy_match: whether to use fuzzy string matching for comparing
      non-categorical slot values.

  Returns:
    goal_acc: a dict whose values are average / joint
        all-goal / categorical-goal / non-categorical-goal accuracies.
  """
    goal_acc = {}

    list_acc, slot_active, slot_cat = compare_slot_values(
        frame_ref["state"]["slot_values"], frame_hyp["state"]["slot_values"], service, no_fuzzy_match
    )

    # (4) Average goal accuracy.
    active_acc = [acc for acc, active in zip(list_acc, slot_active) if active]
    goal_acc[AVERAGE_GOAL_ACCURACY] = np.mean(active_acc) if active_acc else NAN_VAL
    # (4-a) categorical.
    active_cat_acc = [acc for acc, active, cat in zip(list_acc, slot_active, slot_cat) if active and cat]
    goal_acc[AVERAGE_CAT_ACCURACY] = np.mean(active_cat_acc) if active_cat_acc else NAN_VAL
    # (4-b) non-categorical.
    active_noncat_acc = [acc for acc, active, cat in zip(list_acc, slot_active, slot_cat) if active and not cat]
    goal_acc[AVERAGE_NONCAT_ACCURACY] = np.mean(active_noncat_acc) if active_noncat_acc else NAN_VAL

    # (5) Joint goal accuracy.
    goal_acc[JOINT_GOAL_ACCURACY] = np.prod(list_acc) if list_acc else NAN_VAL
    # (5-a) categorical.
    cat_acc = [acc for acc, cat in zip(list_acc, slot_cat) if cat]
    goal_acc[JOINT_CAT_ACCURACY] = np.prod(cat_acc) if cat_acc else NAN_VAL
    # (5-b) non-categorical.
    noncat_acc = [acc for acc, cat in zip(list_acc, slot_cat) if not cat]
    goal_acc[JOINT_NONCAT_ACCURACY] = np.prod(noncat_acc) if noncat_acc else NAN_VAL

    return goal_acc



# -*- coding: utf-8 -*-

# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

__all__ = ["skip_in_data_parallel", "run_only_on_device"]

from functools import partial

from nemo.core.neural_factory import DeviceType


def skip_in_data_parallel(cls):
    """ Decorator that adds the skip_in_data_parallel_ property (set to True) to neural module. """

    def decorator():
        setattr(cls, "skip_in_data_parallel", True)
        return cls

    return decorator


def run_only_on_device(cls=None, device_type: "DeviceType" = DeviceType.CPU):
    """ Decorator that adds the run_only_on_device property to neural module.
    Args:
        device_type: Type device to be set.
    """

    if cls is None:
        return partial(run_only_on_device, device_type=device_type)

    def decorator():
        setattr(cls, "run_only_on_device", device_type)
        return cls

    return decorator



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import numpy as np

from nemo import logging
from nemo.collections.nlp.utils import get_vocab

__all__ = ['read_intent_slot_outputs']


def read_intent_slot_outputs(
    queries, intent_file, slot_file, intent_logits, slot_logits, slot_masks, intents=None, slots=None
):
    intent_dict = get_vocab(intent_file)
    slot_dict = get_vocab(slot_file)
    pred_intents = np.argmax(intent_logits, 1)
    pred_slots = np.argmax(slot_logits, axis=2)
    slot_masks = slot_masks > 0.5
    for i, query in enumerate(queries):
        logging.info(f'Query: {query}')
        pred = pred_intents[i]
        logging.info(f'Predicted intent:\t{pred}\t{intent_dict[pred]}')
        if intents is not None:
            logging.info(f'True intent:\t{intents[i]}\t{intent_dict[intents[i]]}')

        pred_slot = pred_slots[i][slot_masks[i]]
        tokens = query.strip().split()

        if len(pred_slot) != len(tokens):
            raise ValueError('Pred_slot and tokens must be of the same length')

        for j, token in enumerate(tokens):
            output = f'{token}\t{slot_dict[pred_slot[j]]}'
            if slots is not None:
                output = f'{output}\t{slot_dict[slots[i][j]]}'
            logging.info(output)



""" These NeuralModules are based on PyTorch's tutorial:
https://pytorch.org/tutorials/beginner/chatbot_tutorial.html
"""
import random
from typing import Iterable, Mapping, Optional

import torch as t
import torch.nn as nn
import torch.nn.functional as F

from .....core import DeviceType
from .....core.neural_types import *
from ...nm import DataLayerNM, LossNM, TrainableNM
from ..chatbot import data
from nemo.utils.decorators import add_port_docs


class DialogDataLayer(DataLayerNM):
    """Class representing data layer for a chatbot."""

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            "src": NeuralType(('T', 'B'), ChannelType()),
            "src_lengths": NeuralType(tuple('B'), LengthsType()),
            "tgt": NeuralType(('T', 'B'), LabelsType()),
            "mask": NeuralType(('T', 'B'), ChannelType()),
            "max_tgt_lengths": NeuralType(axes=None),
        }

    def __init__(self, batch_size, corpus_name, datafile, min_count=3):
        super().__init__()

        self._batch_size = batch_size
        self._corpus_name = corpus_name
        self._datafile = datafile
        self._min_count = min_count
        voc, pairs = data.loadPrepareData(self._corpus_name, self._datafile)
        self.voc = voc
        self.pairs = data.trimRareWords(voc, pairs, self._min_count)

        self._device = t.device("cuda" if self.placement == DeviceType.GPU else "cpu")
        self._dataloader = []
        for i in range(self.__len__()):
            self._dataloader.append(self.__getitem__(i))

    def __len__(self):
        return len(self.pairs) // self._batch_size

    def __getitem__(self, idx):
        return [
            x.to(self._device) if isinstance(x, t.Tensor) else x
            for x in data.batch2TrainData(self.voc, [random.choice(self.pairs) for _ in range(self._batch_size)],)
        ]

    def get_weights(self) -> Iterable[Optional[Mapping]]:
        return None

    @property
    def data_iterator(self):
        return self._dataloader

    @property
    def dataset(self):
        return None


class EncoderRNN(TrainableNM):
    """RNN-based encoder Neural Module
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            "input_seq": NeuralType(('T', 'B'), ChannelType()),
            "input_lengths": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            "outputs": NeuralType(('T', 'B', 'D'), ChannelType()),
            "hidden": NeuralType(('B', 'D'), ChannelType()),
        }

    def __init__(self, voc_size, encoder_n_layers, hidden_size, dropout, bidirectional=True):
        super().__init__()

        self.voc_size = voc_size
        self.n_layers = encoder_n_layers
        self.hidden_size = hidden_size
        self.dropout = dropout
        self.bidirectional = bidirectional

        # Define layers
        self.embedding = nn.Embedding(self.voc_size, self.hidden_size)
        self.embedding_dropout = nn.Dropout(self.dropout)
        # Initialize GRU; the input_size and hidden_size params are both set
        # to 'hidden_size'
        #   because our input size is a word embedding with number of
        #   features == hidden_size
        self.gru = nn.GRU(
            self.hidden_size,
            self.hidden_size,
            self.n_layers,
            dropout=(0 if self.n_layers == 1 else self.dropout),
            bidirectional=self.bidirectional,
        )
        self._device = t.device("cuda" if self.placement == DeviceType.GPU else "cpu")
        self.to(self._device)

    def forward(self, input_seq, input_lengths, hidden=None):
        embedded = self.embedding(input_seq)
        embedded = self.embedding_dropout(embedded)
        # Pack padded batch of sequences for RNN module
        packed = t.nn.utils.rnn.pack_padded_sequence(embedded, input_lengths)
        # Forward pass through GRU
        outputs, hidden = self.gru(packed, hidden)
        # Unpack padding
        outputs, _ = t.nn.utils.rnn.pad_packed_sequence(outputs)
        # Sum bidirectional GRU outputs
        outputs = outputs[:, :, : self.hidden_size] + outputs[:, :, self.hidden_size :]
        # Return output and final hidden state
        return outputs, hidden


class LuongAttnDecoderRNN(TrainableNM):
    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            "targets": NeuralType(('T', 'B'), LabelsType()),
            "encoder_outputs": NeuralType(('T', 'B', 'D'), ChannelType()),
            "max_target_len": NeuralType(axes=None),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        outputs:
            0: AxisType(TimeTag)

            1: AxisType(BatchTag)

            2: AxisType(ChannelTag)

        hidden:
            0: AxisType(BatchTag)

            1: AxisType(ChannelTag)
        """
        return {
            "outputs": NeuralType(('T', 'B', 'D'), ChannelType()),
            "hidden": NeuralType(('B', 'D'), ChannelType()),
        }

    def __init__(self, attn_model, hidden_size, voc_size, decoder_n_layers, dropout):
        super().__init__()

        self.attn_model = attn_model
        self.hidden_size = hidden_size
        self.voc_size = voc_size
        # output_size
        self.output_size = voc_size
        self.n_layers = decoder_n_layers
        self.dropout = dropout

        # Define layers
        self.embedding = nn.Embedding(self.voc_size, self.hidden_size)
        self.embedding_dropout = nn.Dropout(self.dropout)
        self.gru = nn.GRU(
            self.hidden_size, self.hidden_size, self.n_layers, dropout=(0 if self.n_layers == 1 else self.dropout),
        )
        self.concat = nn.Linear(self.hidden_size * 2, self.hidden_size)
        self.out = nn.Linear(self.hidden_size, self.output_size)

        # Luong attention layer
        class Attn(t.nn.Module):
            def __init__(self, method, hidden_size):
                super(Attn, self).__init__()
                self.method = method
                if self.method not in ["dot", "general", "concat"]:
                    raise ValueError(self.method, "is not an appropriate attention method.")
                self.hidden_size = hidden_size
                if self.method == "general":
                    self.attn = t.nn.Linear(self.hidden_size, hidden_size)
                elif self.method == "concat":
                    self.attn = t.nn.Linear(self.hidden_size * 2, hidden_size)
                    self.v = t.nn.Parameter(t.FloatTensor(hidden_size))

            def dot_score(self, hidden, encoder_output):
                return t.sum(hidden * encoder_output, dim=2)

            def general_score(self, hidden, encoder_output):
                energy = self.attn(encoder_output)
                return t.sum(hidden * energy, dim=2)

            def concat_score(self, hidden, encoder_output):
                energy = self.attn(t.cat((hidden.expand(encoder_output.size(0), -1, -1), encoder_output,), 2,)).tanh()
                return t.sum(self.v * energy, dim=2)

            def forward(self, hidden, encoder_outputs):
                # Calculate the attention weights (energies) based on the
                # given method
                if self.method == "general":
                    attn_energies = self.general_score(hidden, encoder_outputs)
                elif self.method == "concat":
                    attn_energies = self.concat_score(hidden, encoder_outputs)
                elif self.method == "dot":
                    attn_energies = self.dot_score(hidden, encoder_outputs)

                # Transpose max_length and batch_size dimensions
                attn_energies = attn_energies.t()

                # Return the softmax normalized probability scores (with
                # added dimension)
                return F.softmax(attn_energies, dim=1).unsqueeze(1)

        self.attn = Attn(self.attn_model, self.hidden_size)

        self._device = t.device("cuda" if self.placement == DeviceType.GPU else "cpu")
        self.to(self._device)

    def one_step_forward(self, embedded, last_hidden, encoder_outputs):
        embedded = self.embedding_dropout(embedded)
        # Forward through unidirectional GRU
        rnn_output, hidden = self.gru(embedded, last_hidden)
        # Calculate attention weights from the current GRU output
        attn_weights = self.attn(rnn_output, encoder_outputs)
        # Multiply attention weights to encoder outputs to get new "weighted
        # sum" context vector
        context = attn_weights.bmm(encoder_outputs.transpose(0, 1))
        # Concatenate weighted context vector and GRU output using Luong eq. 5
        rnn_output = rnn_output.squeeze(0)
        context = context.squeeze(1)
        concat_input = t.cat((rnn_output, context), 1)
        concat_output = t.tanh(self.concat(concat_input))
        # Predict next word using Luong eq. 6
        output = self.out(concat_output)
        output = F.softmax(output, dim=1)
        # Return output and final hidden state
        return output, hidden

    def forward(self, targets, encoder_outputs, max_target_len):
        SOS_token = 1  # Start-of-sentence token

        decoder_input = t.LongTensor([[SOS_token for _ in range(encoder_outputs.shape[1])]])
        decoder_input = decoder_input.to(self._device)
        decoder_hidden = None
        decoder_output = []
        for step_t in range(max_target_len):
            decoder_inpt_embd = self.embedding(decoder_input)
            decoder_step_output, decoder_hidden = self.one_step_forward(
                embedded=decoder_inpt_embd, last_hidden=decoder_hidden, encoder_outputs=encoder_outputs,
            )
            decoder_output.append(decoder_step_output)
            # Teacher forcing: next input is current target
            decoder_input = targets[step_t].view(1, -1)
        result = t.stack(decoder_output, dim=0)
        return result, decoder_hidden


class MaskedXEntropyLoss(LossNM):
    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            "predictions": NeuralType(('T', 'B', 'D'), ChannelType()),
            "target": NeuralType(('T', 'B'), LabelsType()),
            "mask": NeuralType(('T', 'B'), ChannelType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        loss:
            NeuralType(None)
        """
        return {"loss": NeuralType(axes=None, elements_type=LossType())}

    def __init__(self):
        super().__init__()

        self._device = t.device("cuda" if self.placement == DeviceType.GPU else "cpu")

    def _loss(self, inp, target, mask):
        inp = inp.view(-1, inp.shape[2])
        mask = mask.view(-1, 1)
        crossEntropy = -t.log(t.gather(inp, 1, target.view(-1, 1)))
        loss = crossEntropy.masked_select(mask).mean()
        loss = loss.to(self._device)
        return loss

    def _loss_function(self, **kwargs):
        return self._loss(*(kwargs.values()))


class GreedyLuongAttnDecoderRNN(TrainableNM):
    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {"encoder_outputs": NeuralType(('T', 'B', 'D'), ChannelType())}

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            "outputs": NeuralType(('T', 'B'), ChannelType()),
            "hidden": NeuralType(('B', 'D'), ChannelType()),
        }

    def __init__(self, attn_model, hidden_size, voc_size, decoder_n_layers, dropout, max_dec_steps=10):
        super().__init__()

        self.attn_model = attn_model
        self.hidden_size = hidden_size
        self.voc_size = voc_size
        self.output_size = voc_size
        self.n_layers = decoder_n_layers
        self.dropout = dropout
        self.max_decoder_steps = max_dec_steps

        # Define layers
        self.embedding = nn.Embedding(self.voc_size, self.hidden_size)
        self.embedding_dropout = nn.Dropout(self.dropout)
        self.gru = nn.GRU(
            self.hidden_size, self.hidden_size, self.n_layers, dropout=(0 if self.n_layers == 1 else self.dropout),
        )
        self.concat = nn.Linear(self.hidden_size * 2, self.hidden_size)
        self.out = nn.Linear(self.hidden_size, self.output_size)

        # Luong attention layer
        class Attn(t.nn.Module):
            def __init__(self, method, hidden_size):
                super(Attn, self).__init__()
                self.method = method
                if self.method not in ["dot", "general", "concat"]:
                    raise ValueError(self.method, "is not an appropriate attention method.")
                self.hidden_size = hidden_size
                if self.method == "general":
                    self.attn = t.nn.Linear(self.hidden_size, hidden_size)
                elif self.method == "concat":
                    self.attn = t.nn.Linear(self.hidden_size * 2, hidden_size)
                    self.v = t.nn.Parameter(t.FloatTensor(hidden_size))

            def dot_score(self, hidden, encoder_output):
                return t.sum(hidden * encoder_output, dim=2)

            def general_score(self, hidden, encoder_output):
                energy = self.attn(encoder_output)
                return t.sum(hidden * energy, dim=2)

            def concat_score(self, hidden, encoder_output):
                energy = self.attn(t.cat((hidden.expand(encoder_output.size(0), -1, -1), encoder_output,), 2,)).tanh()
                return t.sum(self.v * energy, dim=2)

            def forward(self, hidden, encoder_outputs):
                # Calculate the attention weights (energies) based on the
                # given method
                if self.method == "general":
                    attn_energies = self.general_score(hidden, encoder_outputs)
                elif self.method == "concat":
                    attn_energies = self.concat_score(hidden, encoder_outputs)
                elif self.method == "dot":
                    attn_energies = self.dot_score(hidden, encoder_outputs)

                # Transpose max_length and batch_size dimensions
                attn_energies = attn_energies.t()

                # Return the softmax normalized probability scores (with
                # added dimension)
                return F.softmax(attn_energies, dim=1).unsqueeze(1)

        self.attn = Attn(self.attn_model, self.hidden_size)

        self._device = t.device("cuda" if self.placement == DeviceType.GPU else "cpu")
        self.to(self._device)

    def one_step_forward(self, embedded, last_hidden, encoder_outputs):
        embedded = self.embedding_dropout(embedded)
        # Forward through unidirectional GRU
        rnn_output, hidden = self.gru(embedded, last_hidden)
        # Calculate attention weights from the current GRU output
        attn_weights = self.attn(rnn_output, encoder_outputs)
        # Multiply attention weights to encoder outputs to get new "weighted
        # sum" context vector
        context = attn_weights.bmm(encoder_outputs.transpose(0, 1))
        # Concatenate weighted context vector and GRU output using Luong eq. 5
        rnn_output = rnn_output.squeeze(0)
        context = context.squeeze(1)
        concat_input = t.cat((rnn_output, context), 1)
        concat_output = t.tanh(self.concat(concat_input))
        # Predict next word using Luong eq. 6
        output = self.out(concat_output)
        output = F.softmax(output, dim=1)
        # Return output and final hidden state
        return output, hidden

    def forward(self, encoder_outputs):
        SOS_token = 1  # Start-of-sentence token
        encoder_outputs = encoder_outputs.detach()

        decoder_input = t.LongTensor([[SOS_token for _ in range(encoder_outputs.shape[1])]])
        decoder_input = decoder_input.to(self._device)
        decoder_hidden = None
        decoder_output = []
        done = False
        for step_t in range(self.max_decoder_steps):
            decoder_inpt_embd = self.embedding(decoder_input)
            decoder_step_output, decoder_hidden = self.one_step_forward(
                embedded=decoder_inpt_embd, last_hidden=decoder_hidden, encoder_outputs=encoder_outputs,
            )
            decoder_output.append(decoder_step_output)
            # Teacher forcing: next input is current target
            _, topi = decoder_step_output.topk(1)
            topi = topi.detach()
            # if topi.item() == EOS_token:
            #  break
            decoder_input = t.LongTensor([[topi[i][0] for i in range(topi.shape[0])]])
            decoder_input = decoder_input.to(self._device)
            # decoder_input = targets[step_t].view(1, -1)
        result_logits = t.stack(decoder_output, dim=0)
        _, result = result_logits.topk(1)
        return result.squeeze(-1), decoder_hidden
        # return result, decoder_hidden



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

'''
This file contains code artifacts adapted from the original implementation:
https://github.com/thu-coai/ConvLab-2/
'''
import torch

from nemo.backends.pytorch.nm import NonTrainableNM
from nemo.collections.nlp.neural_types import *
from nemo.core.neural_types import *
from nemo.utils import logging
from nemo.utils.decorators import add_port_docs

__all__ = ['UserUtteranceEncoder']


class UserUtteranceEncoder(NonTrainableNM):
    """
    Updates dialogue history with user utterance and encodes the history (system and user utterances) into 
    a flat list of tokens (per sample).
    Args:
        data_desc (obj): data descriptor for MultiWOZ dataset, contains information about domains, slots, 
            and associated vocabulary
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        user_uttr (str): user utterance
        dialog_history (list): dialogue history, list of system and diaglogue utterances
        """
        return {
            'user_uttr': NeuralType(axes=[AxisType(kind=AxisKind.Batch, is_list=True)], elements_type=UserUtterance()),
            'dialog_history': NeuralType(
                axes=(AxisType(kind=AxisKind.Batch, is_list=True), AxisType(kind=AxisKind.Time, is_list=True),),
                elements_type=AgentUtterance(),
            ),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        dialog_ids (int): token ids for dialogue history
        dialog_lens (int): length of the tokenized dialogue history
        dialog_history (list): dialogue history, being a list of user and system utterances.
        """
        return {
            # TODO: second dimension is not TIME - as it is the concatenation of list of words (ids) of  all utterances!
            'dialog_ids': NeuralType(('B', 'T'), elements_type=TokenIndex()),
            'dialog_lens': NeuralType(tuple('B'), elements_type=Length()),
            'dialog_history': NeuralType(
                axes=(AxisType(kind=AxisKind.Batch, is_list=True), AxisType(kind=AxisKind.Time, is_list=True),),
                elements_type=AgentUtterance(),
            ),
        }

    def __init__(self, data_desc):
        """
        Initializes the object
        Args:
            data_desc (obj): data descriptor for MultiWOZ dataset, contains information about domains, slots, 
                    and associated vocabulary
        """
        super().__init__()
        self.data_desc = data_desc

    def forward(self, user_uttr, dialog_history):
        """
        Returns dialogue utterances in the format accepted by the TRADE Dialogue state tracking model
        Args:
            dialog_history (list): dialogue history, list of system and diaglogue utterances
            user_uttr (str): user utterance
        Returns:
            dialog_ids (int): token ids for the whole dialogue history
            dialog_lens (int): length of the whole tokenized dialogue history
            dialog_history (list): updated dialogue history, list of system and diaglogue utterances
        """
        # TODO: why we update sys utterance, whereas we have only user utterance at that point?
        dialog_history.append(["user", user_uttr])
        logging.debug("Dialogue history: %s", dialog_history)

        context = ' ; '.join([item[1].strip().lower() for item in dialog_history]).strip() + ' ;'
        context_ids = self.data_desc.vocab.tokens2ids(context.split())
        dialog_ids = torch.tensor(context_ids).unsqueeze(0).to(self._device)
        dialog_lens = torch.tensor(len(context_ids)).unsqueeze(0).to(self._device)

        # logging.debug("!! dialog_ids: %s", dialog_ids)
        # logging.debug("!! dialog_lens: %s", dialog_lens)

        return dialog_ids, dialog_lens, dialog_history



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2018 The Google AI Language Team Authors and
# The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from typing import List, Optional

from transformers import (
    ROBERTA_PRETRAINED_CONFIG_ARCHIVE_MAP,
    ROBERTA_PRETRAINED_MODEL_ARCHIVE_LIST,
    RobertaConfig,
    RobertaModel,
)

from nemo.backends.pytorch.nm import TrainableNM
from nemo.core.neural_modules import PretrainedModelInfo
from nemo.core.neural_types import ChannelType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['Roberta']


class Roberta(TrainableNM):
    """
    ROBERTA wraps around the Huggingface implementation of ROBERTA from their
    transformers repository for easy use within NeMo.

    Args:
        pretrained_model_name (str): If using a pretrained model, this should
            be the model's name. Otherwise, should be left as None.
        config_filename (str): path to model configuration file. Optional.
        vocab_size (int): Size of the vocabulary file, if not using a
            pretrained model.
        hidden_size (int): Size of the encoder and pooler layers.
        num_hidden_layers (int): Number of hidden layers in the encoder.
        num_attention_heads (int): Number of attention heads for each layer.
        intermediate_size (int): Size of intermediate layers in the encoder.
        hidden_act (str): Activation function for encoder and pooler layers;
            "gelu", "relu", and "swish" are supported.
        max_position_embeddings (int): The maximum number of tokens in a
        sequence.
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        input_ids: input token ids
        token_type_ids: segment type ids
        attention_mask: attention mask
        """
        return {
            "input_ids": NeuralType(('B', 'T'), ChannelType()),
            "token_type_ids": NeuralType(('B', 'T'), ChannelType()),
            "attention_mask": NeuralType(('B', 'T'), ChannelType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        hidden_states: output embedding 
        """
        return {"hidden_states": NeuralType(('B', 'T', 'D'), ChannelType())}

    def __init__(
        self,
        pretrained_model_name=None,
        config_filename=None,
        vocab_size=None,
        hidden_size=768,
        num_hidden_layers=12,
        num_attention_heads=12,
        intermediate_size=3072,
        hidden_act="gelu",
        max_position_embeddings=512,
    ):
        super().__init__()

        # Check that only one of pretrained_model_name, config_filename, and
        # vocab_size was passed in
        total = 0
        if pretrained_model_name is not None:
            total += 1
        if config_filename is not None:
            total += 1
        if vocab_size is not None:
            total += 1

        if total != 1:
            raise ValueError(
                "Only one of pretrained_model_name, vocab_size, "
                + "or config_filename should be passed into the "
                + "ROBERTA constructor."
            )

        # TK: The following code checks the same once again.
        if vocab_size is not None:
            config = RobertaConfig(
                vocab_size_or_config_json_file=vocab_size,
                vocab_size=vocab_size,
                hidden_size=hidden_size,
                num_hidden_layers=num_hidden_layers,
                num_attention_heads=num_attention_heads,
                intermediate_size=intermediate_size,
                hidden_act=hidden_act,
                max_position_embeddings=max_position_embeddings,
            )
            model = RobertaModel(config)
        elif pretrained_model_name is not None:
            model = RobertaModel.from_pretrained(pretrained_model_name)
        elif config_filename is not None:
            config = RobertaConfig.from_json_file(config_filename)
            model = RobertaModel(config)
        else:
            raise ValueError(
                "Either pretrained_model_name or vocab_size must" + " be passed into the ROBERTA constructor"
            )

        model.to(self._device)

        self.add_module("roberta", model)
        self.config = model.config
        self._hidden_size = model.config.hidden_size

    @property
    def hidden_size(self):
        """
            Property returning hidden size.

            Returns:
                Hidden size.
        """
        return self._hidden_size

    @staticmethod
    def list_pretrained_models() -> Optional[List[PretrainedModelInfo]]:
        pretrained_models = []
        for key in ROBERTA_PRETRAINED_MODEL_ARCHIVE_LIST:
            model_info = PretrainedModelInfo(
                pretrained_model_name=key,
                description="weights by HuggingFace",
                parameters=ROBERTA_PRETRAINED_CONFIG_ARCHIVE_MAP[key],
                location="",
            )
            pretrained_models.append(model_info)
        return pretrained_models

    def forward(self, input_ids, token_type_ids, attention_mask):
        return self.roberta(input_ids, attention_mask=attention_mask)[0]



from nemo.backends.pytorch.common.losses import *
from nemo.backends.pytorch.common.multi_data import *
from nemo.backends.pytorch.common.other import *
from nemo.backends.pytorch.common.parts import *
from nemo.backends.pytorch.common.rnn import *
from nemo.backends.pytorch.common.search import *
from nemo.backends.pytorch.common.zero_data import *



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2019 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
This file contains code artifacts adapted from the original implementation:
https://github.com/google-research/google-research/blob/master/schema_guided_dst/baseline/data_utils.py
"""

import copy
import json
import os
import pickle
import re
from collections import OrderedDict, defaultdict

import numpy as np
import torch

from nemo.collections.nlp.data.datasets.sgd_dataset.input_example import InputExample
from nemo.utils import logging

__all__ = ['FILE_RANGES', 'PER_FRAME_OUTPUT_FILENAME', 'SGDDataProcessor']

FILE_RANGES = {
    "dstc8_single_domain": {"train": range(1, 44), "dev": range(1, 8), "test": range(1, 12)},
    "dstc8_multi_domain": {"train": range(44, 128), "dev": range(8, 21), "test": range(12, 35)},
    "dstc8_all": {"train": range(1, 128), "dev": range(1, 21), "test": range(1, 35)},
    "multiwoz": {"train": range(1, 18), "dev": range(1, 3), "test": range(1, 3)},
    "debug_sample": {"train": range(1, 2), "dev": range(1, 2), "test": range(1, 2)},
    "sgdplus_single": {"train": range(1, 2), "dev": range(1, 2), "test": range(1, 2)},
    "sgdplus_all": {"train": range(1, 3), "dev": range(1, 3), "test": range(1, 3)},
}

# Name of the file containing all predictions and their corresponding frame metrics.
PER_FRAME_OUTPUT_FILENAME = "dialogues_and_metrics.json"


class SGDDataProcessor(object):
    """Data generator for SGD dialogues."""

    def __init__(
        self, task_name, data_dir, dialogues_example_dir, tokenizer, schema_emb_processor, overwrite_dial_files=False,
    ):
        """
        Constructs SGD8DataProcessor
        Args:
            task_name (str): task  name, for  example, "single_domain"
            data_dir (str): path to data directory
            dialogues_example_dir (str): path to  store processed dialogue examples
            tokenizer (Tokenizer): such as NemoBertTokenizer
            schema_emb_processor (Obj): contains information about schemas
            overwrite_dial_files (bool): whether to overwite dialogue files
        """
        self.data_dir = data_dir
        self.dialogues_examples_dir = dialogues_example_dir

        self._task_name = task_name
        self.schema_config = schema_emb_processor.schema_config
        self.schema_emb_processor = schema_emb_processor

        train_file_range = FILE_RANGES[task_name]["train"]
        dev_file_range = FILE_RANGES[task_name]["dev"]
        test_file_range = FILE_RANGES[task_name]["test"]

        self._file_ranges = {
            "train": train_file_range,
            "dev": dev_file_range,
            "test": test_file_range,
        }

        self._seen_services = {
            "train": set(),
            "dev": set(),
            "test": set(),
        }

        self._tokenizer = tokenizer
        self._max_seq_length = self.schema_config["MAX_SEQ_LENGTH"]

        self.dial_files = {}

        # slots_relation_list.np would contain the candidate list of slots for each (service, slot) which would be
        # looked into when a switch between two services happens in the dialogue and we can not find any value for a slot in the current user utterance.
        # This file would get generated from the dialogues in the training set.
        self.slots_relation_file = os.path.join(dialogues_example_dir, f"{task_name}_train_slots_relation_list.np")

        master_device = not torch.distributed.is_initialized() or torch.distributed.get_rank() == 0
        for dataset in ["train", "dev", "test"]:
            # Process dialogue files
            dial_file = f"{task_name}_{dataset}_examples.processed"
            dial_file = os.path.join(dialogues_example_dir, dial_file)
            self.dial_files[(task_name, dataset)] = dial_file

            dialog_paths = SGDDataProcessor.get_dialogue_files(data_dir, dataset, task_name)
            dialogs = SGDDataProcessor.load_dialogues(dialog_paths)
            for dialog in dialogs:
                self._seen_services[dataset].update(set(dialog['services']))

            if not os.path.exists(dial_file) or overwrite_dial_files:
                logging.debug(f"Start generating the dialogue examples for {dataset} dataset.")
                if master_device:
                    if not os.path.exists(dialogues_example_dir):
                        os.makedirs(dialogues_example_dir)
                    dial_examples, slots_relation_list = self._generate_dialog_examples(
                        dataset, schema_emb_processor.schemas
                    )
                    with open(dial_file, "wb") as f:
                        np.save(f, dial_examples)

                    if dataset == "train":
                        with open(self.slots_relation_file, "wb") as f:
                            pickle.dump(slots_relation_list, f)
                        logging.debug(
                            f"The slot carry-over list for train set is stored at {self.slots_relation_file}"
                        )

                    logging.debug(f"The dialogue examples for {dataset} dataset saved at {dial_file}")
                logging.debug(f"Finish generating the dialogue examples for {dataset} dataset.")

            # wait until the master process writes to the dialogue processed file
            if torch.distributed.is_initialized():
                torch.distributed.barrier()

    def get_dialog_examples(self, dataset):
        """
        Returns a list of `InputExample`s of the data splits' dialogues.
        Args:
          dataset(str): can be "train", "dev", or "test".
        Returns:
          examples: a list of `InputExample`s.
        """
        if (self._task_name, dataset) not in self.dial_files or not os.path.exists(
            self.dial_files[(self._task_name, dataset)]
        ):
            raise ValueError(
                f"{dataset} dialogue examples were not processed for {self._task_name} task. Re-initialize SGDDataProcessor and add {dataset} dataset to datasets arg."
            )
        dial_file = self.dial_files[(self._task_name, dataset)]
        logging.info(f"Loading dialogue examples from {dial_file}.")

        with open(dial_file, "rb") as f:
            dial_examples = np.load(f, allow_pickle=True)

        if not os.path.exists(self.slots_relation_file):
            raise ValueError(
                f"Slots relation file {self.slots_relation_file} does not exist. It is needed for the carry-over mechanism of state tracker for switches between services."
            )

        with open(self.slots_relation_file, "rb") as f:
            self.schema_emb_processor.update_slots_relation_list(pickle.load(f))
        logging.info(
            f"Loaded the slot relation list for value carry-over between services from {self.slots_relation_file}."
        )

        return dial_examples

    def get_seen_services(self, dataset_split):
        return self._seen_services[dataset_split]

    def _generate_dialog_examples(self, dataset, schemas):
        """
        Returns a list of `InputExample`s of the data splits' dialogues.
        Args:
          dataset(str): can be "train", "dev", or "test".
          schemas(Schema): for all services and all datasets processed by the schema_processor
        Returns:
          examples: a list of `InputExample`s.
        """
        logging.info(f'Creating examples and slot relation list from the dialogues started...')
        dialog_paths = [
            os.path.join(self.data_dir, dataset, "dialogues_{:03d}.json".format(i)) for i in self._file_ranges[dataset]
        ]
        dialogs = SGDDataProcessor.load_dialogues(dialog_paths)

        examples = []
        slot_carryover_candlist = defaultdict(int)
        services_switch_counts = defaultdict(int)

        for dialog_idx, dialog in enumerate(dialogs):
            if dialog_idx % 1000 == 0:
                logging.info(f'Processed {dialog_idx} dialogues.')
            examples.extend(
                self._create_examples_from_dialog(
                    dialog, schemas, dataset, slot_carryover_candlist, services_switch_counts
                )
            )

        slots_relation_list = defaultdict(list)
        for slots_relation, relation_size in slot_carryover_candlist.items():
            if relation_size > 0:
                switch_counts = (
                    services_switch_counts[(slots_relation[0], slots_relation[2])]
                    + services_switch_counts[(slots_relation[2], slots_relation[0])]
                )
                relation_size = relation_size / switch_counts
                slots_relation_list[(slots_relation[0], slots_relation[1])].append(
                    (slots_relation[2], slots_relation[3], relation_size)
                )
                slots_relation_list[(slots_relation[2], slots_relation[3])].append(
                    (slots_relation[0], slots_relation[1], relation_size)
                )

        return examples, slots_relation_list

    def _create_examples_from_dialog(self, dialog, schemas, dataset, slot_carryover_candlist, services_switch_counts):
        """
        Create examples for every turn in the dialog.
        Args:
            dialog (dict): dialogue example
            schemas(Schema): for all services and all datasets processed by the schema_processor
            dataset(str): can be "train", "dev", or "test".
            slot_carryover_candlist(dict): a dictionary to keep and count the number of carry-over cases between two slots from two different services
        Returns:
            examples: a list of `InputExample`s.
        """
        dialog_id = dialog["dialogue_id"]
        prev_states = {}
        examples = []
        agg_sys_states = defaultdict(dict)
        agg_sys_states_prev = defaultdict(dict)
        frame_service_prev = ""
        for turn_idx, turn in enumerate(dialog["turns"]):
            # Generate an example for every frame in every user turn.
            if turn["speaker"] == "SYSTEM":
                agg_sys_states_prev = copy.deepcopy(agg_sys_states)
                for frame in turn["frames"]:
                    for action in frame["actions"]:
                        if action["slot"] and len(action["values"]) > 0:
                            agg_sys_states[frame["service"]][action["slot"]] = action["values"]

            if turn["speaker"] == "USER":
                user_utterance = turn["utterance"]
                user_frames = {f["service"]: f for f in turn["frames"]}
                if turn_idx > 0:
                    system_turn = dialog["turns"][turn_idx - 1]
                    system_utterance = system_turn["utterance"]
                    system_frames = {f["service"]: f for f in system_turn["frames"]}
                else:
                    system_utterance = ""
                    system_frames = {}

                if len(user_frames) == 2:
                    frames_list_name = list(user_frames.keys())
                    frames_list_val = list(user_frames.values())
                    user_frames_ordered = OrderedDict()

                    if frame_service_prev != "" and frames_list_name[0] != frame_service_prev:
                        user_frames_ordered[frames_list_name[1]] = frames_list_val[1]
                        user_frames_ordered[frames_list_name[0]] = frames_list_val[0]
                        user_frames = user_frames_ordered

                turn_id = "{}-{}-{:02d}".format(dataset, dialog_id, turn_idx)
                turn_examples, prev_states = self._create_examples_from_turn(
                    turn_id,
                    system_utterance,
                    user_utterance,
                    system_frames,
                    user_frames,
                    prev_states,
                    schemas,
                    copy.deepcopy(agg_sys_states_prev),
                    slot_carryover_candlist,
                    services_switch_counts,
                )
                examples.extend(turn_examples)
                frame_service_prev = user_frames[list(user_frames.keys())[-1]]["service"]
        return examples

    def _get_state_update(self, current_state, prev_state):
        """
        Updates dialogue state
        Args:
            current_state (dict): dict of slot - slot values pairs for the current dialogue turn
            prev_state (dict): dict of slot - slot values pairs for the previous dialogue turns
        Returns:
            state_update (dict): dict of slot - slot values pairs that very added/updated during the current
                dialogue turn
        """
        state_update = dict(current_state)
        for slot, values in current_state.items():
            if slot in prev_state and prev_state[slot][0] in values:
                # Remove the slot from state if its value didn't change.
                state_update.pop(slot)
        return state_update

    def _create_examples_from_turn(
        self,
        turn_id,
        system_utterance,
        user_utterance,
        system_frames,
        user_frames,
        prev_states,
        schemas,
        agg_sys_states,
        slot_carryover_candlist,
        services_switch_counts,
    ):
        """
        Creates an example for each frame in the user turn.
        Args:
            turn_id (int): turn number
            system_utterance (str): last system utterance
            user_utterance (str): lst user utterance
            system_frames (dict): all system utterances and slot - slot value pairs
            user_frames (dict): all user utterances and slot - slot value pairs
            prev_states (dict): slot - slot value pairs from the previous turns
            schemas (obj): carries information about the service from the current turn
            agg_sys_states (dict): the collection of all the slots and values mentioned by the system until the previous turn
            slot_carryover_candlist (dict): a dictionary to keep and aggregate the counts of the relations found between any two slots
            services_switch_counts (dict): a dictionary to keep and aggregate the number of switches between any two services
        Returns:
            examples: a list of `InputExample`s.
            prev_states (dict): updated dialogue state
        """
        system_tokens, system_alignments, system_inv_alignments = self._tokenize(system_utterance)
        user_tokens, user_alignments, user_inv_alignments = self._tokenize(user_utterance)
        states = {}
        base_example = InputExample(
            schema_config=self.schema_config, is_real_example=True, tokenizer=self._tokenizer, service_schema=schemas
        )
        base_example.example_id = turn_id

        _, dialog_id, turn_id_ = turn_id.split('-')
        dialog_id_1, dialog_id_2 = dialog_id.split('_')
        base_example.example_id_num = [int(dialog_id_1), int(dialog_id_2), int(turn_id_)]
        base_example.add_utterance_features(
            system_tokens, system_inv_alignments, user_tokens, user_inv_alignments, system_utterance, user_utterance
        )

        examples = []
        for service, user_frame in user_frames.items():
            # Create an example for this service.
            example = base_example.make_copy_with_utterance_features()

            example.example_id = "{}-{}".format(turn_id, service)
            _, dialog_id, turn_id_ = turn_id.split('-')
            dialog_id_1, dialog_id_2 = dialog_id.split('_')
            example.example_id_num = [
                int(dialog_id_1),
                int(dialog_id_2),
                int(turn_id_),
                schemas.get_service_id(service),
            ]

            example.service_schema = schemas.get_service_schema(service)
            system_frame = system_frames.get(service, None)
            state = user_frame["state"]["slot_values"]
            state_update = self._get_state_update(state, prev_states.get(service, {}))
            states[service] = state

            # Populate features in the example.
            example.add_categorical_slots(state_update, agg_sys_states[service])
            # The input tokens to bert are in the format [CLS] [S1] [S2] ... [SEP]
            # [U1] [U2] ... [SEP] [PAD] ... [PAD]. For system token indices a bias of
            # 1 is added for the [CLS] token and for user tokens a bias of 2 +
            # len(system_tokens) is added to account for [CLS], system tokens and
            # [SEP].
            user_span_boundaries = self._find_subword_indices(
                state_update, user_utterance, user_frame["slots"], user_alignments, user_tokens, 2 + len(system_tokens)
            )
            if system_frame is not None:
                system_span_boundaries = self._find_subword_indices(
                    state_update, system_utterance, system_frame["slots"], system_alignments, system_tokens, 1
                )
            else:
                system_span_boundaries = {}
            example.add_noncategorical_slots(state_update, user_span_boundaries, system_span_boundaries)
            example.add_requested_slots(user_frame)
            example.add_intents(user_frame)
            examples.append(example)

            if service not in prev_states and int(turn_id_) > 0:
                prev_service = ""
                for prev_s, prev_slot_value_list in prev_states.items():
                    if prev_s != service:
                        prev_service = prev_s
                services_switch_counts[(prev_service, service)] += 1

                if prev_service in states:
                    prev_slot_value_list = states[prev_service]
                else:
                    prev_slot_value_list = prev_states[prev_service]

                cur_slot_value_list = state_update
                for cur_slot, cur_values in cur_slot_value_list.items():
                    for prev_slot, prev_values in prev_slot_value_list.items():
                        if "True" in cur_values or "False" in cur_values:
                            continue
                        if set(cur_values) & set(prev_values):
                            slot_carryover_candlist[(prev_service, prev_slot, service, cur_slot)] += 1.0

        return examples, states

    def _find_subword_indices(self, slot_values, utterance, char_slot_spans, alignments, subwords, bias):
        """Find indices for subwords corresponding to slot values."""
        span_boundaries = {}
        for slot, values in slot_values.items():
            # Get all values present in the utterance for the specified slot.
            value_char_spans = {}
            for slot_span in char_slot_spans:
                if slot_span["slot"] == slot:
                    value = utterance[slot_span["start"] : slot_span["exclusive_end"]]
                    start_tok_idx = alignments[slot_span["start"]]
                    end_tok_idx = alignments[slot_span["exclusive_end"] - 1]
                    if 0 <= start_tok_idx < len(subwords):
                        end_tok_idx = min(end_tok_idx, len(subwords) - 1)
                        value_char_spans[value] = (start_tok_idx + bias, end_tok_idx + bias)
            for v in values:
                if v in value_char_spans:
                    span_boundaries[slot] = value_char_spans[v]
                    break
        return span_boundaries

    def _tokenize(self, utterance):
        """Tokenize the utterance using word-piece tokenization used by BERT.

        Args:
          utterance: A string containing the utterance to be tokenized.

        Returns:
          bert_tokens: A list of tokens obtained by word-piece tokenization of the
            utterance.
          alignments: A dict mapping indices of characters corresponding to start
            and end positions of words (not subwords) to corresponding indices in
            bert_tokens list.
          inverse_alignments: A list of size equal to bert_tokens. Each element is a
            tuple containing the index of the starting and inclusive ending
            character of the word corresponding to the subword. This list is used
            during inference to map word-piece indices to spans in the original
            utterance.
        """
        # utterance = tokenization.convert_to_unicode(utterance)

        # After _naive_tokenize, spaces and punctuation marks are all retained, i.e.
        # direct concatenation of all the tokens in the sequence will be the
        # original string.
        tokens = SGDDataProcessor._naive_tokenize(utterance)
        # Filter out empty tokens and obtain aligned character index for each token.
        alignments = {}
        char_index = 0
        bert_tokens = []
        # These lists store inverse alignments to be used during inference.
        bert_tokens_start_chars = []
        bert_tokens_end_chars = []
        for token in tokens:
            if token.strip():
                subwords = self._tokenizer.text_to_tokens(token)
                # Store the alignment for the index of starting character and the
                # inclusive ending character of the token.
                alignments[char_index] = len(bert_tokens)
                bert_tokens_start_chars.extend([char_index] * len(subwords))
                bert_tokens.extend(subwords)
                # The inclusive ending character index corresponding to the word.
                inclusive_char_end = char_index + len(token) - 1
                alignments[inclusive_char_end] = len(bert_tokens) - 1
                bert_tokens_end_chars.extend([inclusive_char_end] * len(subwords))
            char_index += len(token)
        inverse_alignments = list(zip(bert_tokens_start_chars, bert_tokens_end_chars))
        return bert_tokens, alignments, inverse_alignments

    def get_num_dialog_examples(self, dataset):
        """
        Gets the number of dilaog examples in the data split.
        Args:
          dataset: str. can be "train", "dev", or "test".
        Returns:from nemo_nlp.data.datasets.sgd import data_utils
          example_count: int. number of examples in the specified dataset.
        """
        example_count = 0
        dialog_paths = [
            os.path.join(self.data_dir, dataset, "dialogues_{:03d}.json".format(i)) for i in self._file_ranges[dataset]
        ]
        dst_set = SGDDataProcessor.load_dialogues(dialog_paths)
        for dialog in dst_set:
            for turn in dialog["turns"]:
                if turn["speaker"] == "USER":
                    example_count += len(turn["frames"])
        return example_count

    @classmethod
    def _naive_tokenize(cls, s):
        """
        Tokenizes a string, separating words, spaces and punctuations.
        Args:
            s (str): a string
        Returns:
            seq_tok (list): list of words, spaces and punctuations from the s
        """
        # Spaces and punctuation marks are all retained, i.e. direct concatenation
        # of all the tokens in the sequence will be the original string.
        seq_tok = [tok for tok in re.split(r"([^a-zA-Z0-9])", s) if tok]
        return seq_tok

    @classmethod
    def load_dialogues(cls, dialog_json_filepaths):
        """
        Obtain the list of all dialogues from specified json files.
        Args:
            dialog_json_filepaths (list): list of json files
        Returns:
            dialogs  (list): the list of all dialogues
        """
        dialogs = []
        for dialog_json_filepath in sorted(dialog_json_filepaths):
            with open(dialog_json_filepath, 'r') as f:
                dialogs.extend(json.load(f))
                f.close()
        return dialogs

    @classmethod
    def get_dialogue_files(cls, data_dir, dataset_split, task_name):
        """
        Obtain the list of all dialogue json files
        Args:
            data_dir (str): path to the data folde
            dataset_split (str): dev, test or train
            task_name (str): SGD task name, see keys of the FILE_RANGES
        Returns:
            dialogs (list): the list of all dialogue json files paths
        """
        return [
            os.path.join(data_dir, dataset_split, 'dialogues_{:03d}.json'.format(fid))
            for fid in FILE_RANGES[task_name][dataset_split]
        ]



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2019, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nemo.core.neural_types import AxisKindAbstract, ElementType, StringType

__all__ = [
    'DialogAxisKind',
    'Utterance',
    'UserUtterance',
    'SystemUtterance',
    'AgentUtterance',
    'SlotValue',
    'MultiWOZBeliefState',
]


class DialogAxisKind(AxisKindAbstract):
    """ Class containing definitions of axis kinds specialized for the dialog problem domain. """

    Domain = 7


class Utterance(StringType):
    """Element type representing an utterance (e.g. "Is there a train from Ely to Cambridge on Tuesday ?")."""


class UserUtterance(Utterance):
    """Element type representing a utterance expresesd by the user."""


class SystemUtterance(Utterance):
    """Element type representing an utterance produced by the system."""


class AgentUtterance(ElementType):
    """Element type representing utterance returned by an agent (user or system) participating in a dialog."""

    # def __str__(self):
    #    return "Utterance returned by an agent (user or system) participating in a dialog."

    # def fields(self):
    #    return ("agent", "utterance")


class SlotValue(ElementType):
    """Element type representing slot-value pair."""

    # def __str__(self):
    #    return "Slot-value pair"

    # def fields(self):
    #    return ("slot", "value")


class MultiWOZBeliefState(SlotValue):
    """Element type representing MultiWOZ belief state - one per domain."""

    # def fields(self):
    #    return ("book", "semi")



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2019 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Extract BERT embeddings for slots, values, intents in schema.

This file contains code artifacts adapted from the original implementation:
https://github.com/google-research/google-research/blob/master/schema_guided_dst/baseline/extract_schema_embedding.py
"""

import collections
import random
import re

import numpy as np
import torch
from torch.utils.data import Dataset

from nemo import logging
from nemo.collections.nlp.data.datasets.sgd_dataset.input_example import truncate_seq_pair

# Separator to separate the two sentences in BERT's input sequence.
_NL_SEPARATOR = "|||"

__all__ = ['SchemaEmbeddingDataset']


class SchemaEmbeddingDataset(Dataset):
    def __init__(self, schema_config, tokenizer, schemas):
        """Generate the embeddings for a schema's elements.

        Args:
          tokenizer (tokenizer): such as NemoBertTokenizer
          max_seq_length: Sequence length used for BERT model
          schemas: Schemas for all services in the datasets
        """
        self._tokenizer = tokenizer
        self.schema_config = schema_config
        self.schemas = schemas

        input_features = self._get_input_features()

        self.features = collections.defaultdict(list)

        for feature in input_features:
            self.features["input_ids"].append(feature.input_ids)
            self.features["input_mask"].append(feature.input_mask)
            self.features["input_type_ids"].append(feature.input_type_ids)
            self.features["embedding_tensor_name"].append(feature.embedding_tensor_name)
            self.features["service_id"].append(feature.service_id)
            self.features["intent_or_slot_id"].append(feature.intent_or_slot_id)
            self.features["value_id"].append(feature.value_id)

    def __len__(self):
        return len(self.features['input_ids'])

    def __getitem__(self, idx):
        return (
            np.array(self.features['input_ids'][idx]),
            np.array(self.features['input_mask'][idx], dtype=np.long),
            np.array(self.features['input_type_ids'][idx]),
        )

    def _create_feature(self, line, embedding_tensor_name, service_id, intent_or_slot_id, value_id=-1):
        """Create a single InputFeatures instance."""
        seq_length = self.schema_config["MAX_SEQ_LENGTH"]
        # line = tokenization.convert_to_unicode(input_line)
        line = line.strip()
        text_a = None
        text_b = None
        m = re.match(r"^(.*) \|\|\| (.*)$", line)
        if m is None:
            text_a = line
        else:
            text_a = m.group(1)
            text_b = m.group(2)

        tokens_a = self._tokenizer.text_to_tokens(text_a)
        tokens_b = None
        if text_b:
            tokens_b = self._tokenizer.text_to_tokens(text_b)

        if tokens_b:
            # Modifies `tokens_a` and `tokens_b` in place so that the total
            # length is less than the specified length.
            # Account for [CLS], [SEP], [SEP] with "- 3"
            truncate_seq_pair(tokens_a, tokens_b, seq_length - 3)
        else:
            # Account for [CLS] and [SEP] with "- 2"
            if len(tokens_a) > seq_length - 2:
                tokens_a = tokens_a[0 : (seq_length - 2)]

        # The convention in BERT is:
        # (a) For sequence pairs:
        #  tokens:   [CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]
        #  type_ids: 0     0  0    0    0     0       0 0     1  1  1  1   1 1
        # (b) For single sequences:
        #  tokens:   [CLS] the dog is hairy . [SEP]
        #  type_ids: 0     0   0   0  0     0 0
        #
        # Where "type_ids" are used to indicate whether this is the first
        # sequence or the second sequence. The embedding vectors for `type=0` and
        # `type=1` were learned during pre-training and are added to the wordpiece
        # embedding vector (and position vector). This is not *strictly* necessary
        # since the [SEP] token unambiguously separates the sequences, but it
        # makes it easier for the model to learn the concept of sequences.
        #
        # For classification tasks, the first vector (corresponding to [CLS]) is
        # used as as the "sentence vector". Note that this only makes sense
        # because the entire model is fine-tuned.
        tokens = []
        input_type_ids = []
        tokens.append(self._tokenizer.cls_token)
        input_type_ids.append(0)
        for token in tokens_a:
            tokens.append(token)
            input_type_ids.append(0)
        tokens.append(self._tokenizer.sep_token)
        input_type_ids.append(0)

        if tokens_b:
            for token in tokens_b:
                tokens.append(token)
                input_type_ids.append(1)
            tokens.append(self._tokenizer.sep_token)
            input_type_ids.append(1)

        input_ids = self._tokenizer.tokens_to_ids(tokens)

        # The mask has 1 for real tokens and 0 for padding tokens. Only real
        # tokens are attended to.
        input_mask = [1] * len(input_ids)

        # Zero-pad up to the sequence length.
        while len(input_ids) < seq_length:
            input_ids.append(0)
            input_mask.append(0)
            input_type_ids.append(0)
        assert len(input_ids) == seq_length
        assert len(input_mask) == seq_length
        assert len(input_type_ids) == seq_length

        return InputFeatures(
            input_ids=input_ids,
            input_mask=input_mask,
            input_type_ids=input_type_ids,
            embedding_tensor_name=embedding_tensor_name,
            service_id=service_id,
            intent_or_slot_id=intent_or_slot_id,
            value_id=value_id,
        )

    def _get_intents_input_features(self, service_schema):
        """Create features for BERT inference for all intents of a service.

      We use "[service description] ||| [intent name] [intent description]" as an
        intent's full description.

      Args:
        service_schema: A ServiceSchema object containing the schema for the
          corresponding service.

      Returns:
        A list of InputFeatures containing features to be given as input to the
        BERT model.
      """
        service_des = service_schema.description

        features = []
        intent_descriptions = {i["name"]: i["description"] for i in service_schema.schema_json["intents"]}
        for intent_id, intent in enumerate(service_schema.intents):
            nl_seq = " ".join([service_des, _NL_SEPARATOR, intent, intent_descriptions[intent]])
            features.append(self._create_feature(nl_seq, "intent_emb", service_schema.service_id, intent_id))
        return features

    def _get_req_slots_input_features(self, service_schema):
        """Create features for BERT inference for all requested slots of a service.

      We use "[service description] ||| [slot name] [slot description]" as a
        slot's full description.

      Args:
        service_schema: A ServiceSchema object containing the schema for the
          corresponding service.

      Returns:
        A list of InputFeatures containing features to be given as input to the
        BERT model.
      """
        service_des = service_schema.description

        slot_descriptions = {s["name"]: s["description"] for s in service_schema.schema_json["slots"]}
        features = []
        for slot_id, slot in enumerate(service_schema.slots):
            nl_seq = " ".join([service_des, _NL_SEPARATOR, slot, slot_descriptions[slot]])
            features.append(self._create_feature(nl_seq, "req_slot_emb", service_schema.service_id, slot_id))
        return features

    def _get_goal_slots_and_values_input_features(self, service_schema):
        """Get BERT input features for all goal slots and categorical values.

      We use "[service description] ||| [slot name] [slot description]" as a
        slot's full description.
      We use ""[slot name] [slot description] ||| [value name]" as a categorical
        slot value's full description.

      Args:
        service_schema: A ServiceSchema object containing the schema for the
          corresponding service.

      Returns:
        A list of InputFeatures containing features to be given as input to the
        BERT model.
      """
        service_des = service_schema.description

        features = []
        slot_descriptions = {s["name"]: s["description"] for s in service_schema.schema_json["slots"]}

        for slot_id, slot in enumerate(service_schema.non_categorical_slots):
            nl_seq = " ".join([service_des, _NL_SEPARATOR, slot, slot_descriptions[slot]])
            features.append(self._create_feature(nl_seq, "noncat_slot_emb", service_schema.service_id, slot_id))

        for slot_id, slot in enumerate(service_schema.categorical_slots):
            nl_seq = " ".join([service_des, _NL_SEPARATOR, slot, slot_descriptions[slot]])
            features.append(self._create_feature(nl_seq, "cat_slot_emb", service_schema.service_id, slot_id))
            for value_id, value in enumerate(service_schema.get_categorical_slot_values(slot)):
                nl_seq = " ".join([slot, slot_descriptions[slot], _NL_SEPARATOR, value])
                features.append(
                    self._create_feature(nl_seq, "cat_slot_value_emb", service_schema.service_id, slot_id, value_id)
                )
        return features

    def _get_input_features(self):
        """Get the input function to compute schema element embeddings.

        Args:
          schemas: A wrapper for all service schemas in the dataset to be embedded.

        Returns:
          The input_fn to be passed to the estimator.
        """
        # Obtain all the features.
        features = []
        for service in self.schemas.services:
            service_schema = self.schemas.get_service_schema(service)
            features.extend(self._get_intents_input_features(service_schema))
            features.extend(self._get_req_slots_input_features(service_schema))
            features.extend(self._get_goal_slots_and_values_input_features(service_schema))

        return features

    def _populate_schema_embeddings(self, schema_embeddings, hidden_states, mode):
        """
        Populate all schema embeddings with BERT embeddings.
        """
        completed_services = set()
        batch_size, seq_len, hidden_size = hidden_states[0].shape

        for idx in range(len(self)):
            service_id = self.features['service_id'][idx]
            service = self.schemas.get_service_from_id(service_id)

            if service not in completed_services:
                logging.debug(f"Generating embeddings for service {service}.")
                completed_services.add(service)
            tensor_name = self.features["embedding_tensor_name"][idx]
            emb_mat = schema_embeddings[service_id][tensor_name]

            if mode == 'random':
                # randomly initialize schema embeddings
                random_token = random.randint(0, seq_len - 1)
                embedding = [round(float(x), 6) for x in hidden_states[0][idx, random_token, :].flat]
            elif mode == 'last_layer_average':
                # Obtain the encoding of the [CLS] token.
                embedding = [round(float(x), 6) for x in np.mean(hidden_states[0][idx, :], 0).flat]
            elif mode == 'baseline':
                # Obtain the encoding of the [CLS] token.
                embedding = [round(float(x), 6) for x in hidden_states[0][idx, 0, :].flat]
            else:
                raise ValueError(f'Mode {mode} for generation schema embeddings is not supported')
            intent_or_slot_id = self.features['intent_or_slot_id'][idx]
            value_id = self.features['value_id'][idx]

            if tensor_name == "cat_slot_value_emb":
                emb_mat[intent_or_slot_id, value_id] = embedding
            else:
                emb_mat[intent_or_slot_id] = embedding

    def save_embeddings(self, bert_hidden_states, output_file, mode):
        """Generate schema element embeddings and save it as a numpy file."""
        schema_embeddings = []
        max_num_intent = self.schema_config["MAX_NUM_INTENT"]
        max_num_cat_slot = self.schema_config["MAX_NUM_CAT_SLOT"]
        max_num_noncat_slot = self.schema_config["MAX_NUM_NONCAT_SLOT"]
        max_num_slot = max_num_cat_slot + max_num_noncat_slot
        max_num_value = self.schema_config["MAX_NUM_VALUE_PER_CAT_SLOT"]
        embedding_dim = self.schema_config["EMBEDDING_DIMENSION"]

        for _ in self.schemas.services:
            schema_embeddings.append(
                {
                    "intent_emb": np.zeros([max_num_intent, embedding_dim]),
                    "req_slot_emb": np.zeros([max_num_slot, embedding_dim]),
                    "cat_slot_emb": np.zeros([max_num_cat_slot, embedding_dim]),
                    "noncat_slot_emb": np.zeros([max_num_noncat_slot, embedding_dim]),
                    "cat_slot_value_emb": np.zeros([max_num_cat_slot, max_num_value, embedding_dim]),
                }
            )

        # Populate the embeddings based on bert inference results and save them.
        self._populate_schema_embeddings(schema_embeddings, bert_hidden_states, mode)

        master_device = not torch.distributed.is_initialized() or torch.distributed.get_rank() == 0
        if master_device:
            with open(output_file, "wb") as f_s:
                np.save(f_s, schema_embeddings)
                logging.info(f"The schema embeddings saved at {output_file}")
                f_s.close()


class InputFeatures(object):
    """A single set of features for BERT inference."""

    def __init__(
        self, input_ids, input_mask, input_type_ids, embedding_tensor_name, service_id, intent_or_slot_id, value_id
    ):
        # The ids in the vocabulary for input tokens.
        self.input_ids = input_ids
        # A boolean mask indicating which tokens in the input_ids are valid.
        self.input_mask = input_mask
        # Denotes the sequence each input token belongs to.
        self.input_type_ids = input_type_ids
        # The name of the embedding tensor corresponding to this example.
        self.embedding_tensor_name = embedding_tensor_name
        # The id of the service corresponding to this example.
        self.service_id = service_id
        # The id of the intent (for intent embeddings) or slot (for slot or slot
        # value embeddings) corresponding to this example.
        self.intent_or_slot_id = intent_or_slot_id
        # The id of the value corresponding to this example. Only set if slot value
        # embeddings are being calculated.
        self.value_id = value_id



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import re
import string

import numpy as np

__all__ = ['get_vocab', 'get_tokens', 'normalize_answer', 'mask_padded_tokens', 'concatenate']


def get_vocab(file):
    lines = open(file, 'r').readlines()
    lines = [line.strip() for line in lines if line.strip()]
    labels = {i: lines[i] for i in range(len(lines))}
    return labels


def normalize_answer(s):
    """Lower text and remove punctuation, articles and extra whitespace."""

    def remove_articles(text):
        return re.sub(r'\b(a|an|the)\b', ' ', text)

    def white_space_fix(text):
        return ' '.join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return ''.join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(s))))


def get_tokens(s):
    if not s:
        return []
    return normalize_answer(s).split()


def mask_padded_tokens(tokens, pad_id):
    mask = tokens != pad_id
    return mask


def concatenate(lists):
    """
    Helper function for inference
    """
    return np.concatenate([t.cpu() for t in lists])



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from transformers import GPT2Tokenizer

from nemo.collections.nlp.data.tokenizers.tokenizer_spec import TokenizerSpec

__all__ = ['NemoGPT2Tokenizer']


class NemoGPT2Tokenizer(TokenizerSpec):
    def __init__(
        self,
        pretrained_model=None,
        vocab_file=None,
        merges_file=None,
        errors='replace',
        bos_token="<|endoftext|>",
        eos_token="<|endoftext|>",
    ):
        if pretrained_model:
            self.tokenizer = GPT2Tokenizer.from_pretrained(pretrained_model)
        self.vocab_size = self.tokenizer.vocab_size
        special_tokens_dict = {}
        if self.tokenizer.unk_token is None:
            self.tokenizer.unk_token = "<|unk|>"
            special_tokens_dict["unk_token"] = "<|unk|>"
        if self.tokenizer.bos_token is None:
            special_tokens_dict["bos_token"] = bos_token
        if self.tokenizer.eos_token is None:
            special_tokens_dict["eos_token"] = eos_token
        if self.tokenizer.pad_token is None:
            special_tokens_dict["pad_token"] = "<|pad|>"
        self.tokenizer.add_special_tokens(special_tokens_dict)

    def text_to_tokens(self, text):
        tokens = self.tokenizer.tokenize(text)
        return tokens

    def tokens_to_text(self, tokens):
        text = self.tokenizer.convert_tokens_to_string(tokens)
        return text

    def tokens_to_ids(self, tokens):
        ids = self.tokenizer.convert_tokens_to_ids(tokens)
        return ids

    def ids_to_tokens(self, ids):
        tokens = self.tokenizer.convert_ids_to_tokens(ids)
        return tokens

    def text_to_ids(self, text):
        tokens = self.text_to_tokens(text)
        ids = self.tokens_to_ids(tokens)
        return ids

    def ids_to_text(self, ids):
        tokens = self.ids_to_tokens(ids)
        text = self.tokens_to_text(tokens)
        return text

    @property
    def pad_id(self):
        return self.tokens_to_ids([self.tokenizer.pad_token])[0]

    @property
    def bos_id(self):
        return self.tokens_to_ids([self.tokenizer.bos_token])[0]

    @property
    def eos_id(self):
        return self.tokens_to_ids([self.tokenizer.eos_token])[0]



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import os

import torch
from megatron.initialize import initialize_megatron
from megatron.model.bert_model import bert_attention_mask_func, bert_extended_attention_mask, bert_position_ids
from megatron.model.language_model import get_language_model
from megatron.model.utils import init_method_normal, scaled_init_method_normal

from nemo.backends.pytorch.nm import TrainableNM
from nemo.core import DeviceType
from nemo.core.neural_types import ChannelType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['MegatronBERT']


class MegatronBERT(TrainableNM):
    """
    MegatronBERT wraps around the Megatron Language model
    from https://github.com/NVIDIA/Megatron-LM

    Args:
        config_file (str): path to model configuration file.
        vocab_file (str): path to vocabulary file.
        tokenizer_type (str): tokenizer type, currently only 'BertWordPieceLowerCase' supported.
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        input_ids: input token ids
        token_type_ids: segment type ids
        attention_mask: attention mask
        """
        return {
            "input_ids": NeuralType(('B', 'T'), ChannelType()),
            "attention_mask": NeuralType(('B', 'T'), ChannelType()),
            "token_type_ids": NeuralType(('B', 'T'), ChannelType(), optional=True),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        hidden_states: output embedding 
        """
        return {"hidden_states": NeuralType(('B', 'T', 'D'), ChannelType())}

    def __init__(
        self,
        model_name,
        vocab_file,
        hidden_size=1024,
        num_attention_heads=16,
        num_layers=24,
        max_seq_length=512,
        tokenizer_type='BertWordPieceLowerCase',
        init_method_std=0.02,
        num_tokentypes=2,
    ):

        super().__init__()

        if not os.path.exists(vocab_file):
            raise ValueError(f'Vocab file not found at {vocab_file}')

        megatron_args = {
            "num_layers": num_layers,
            "hidden_size": hidden_size,
            "num_attention_heads": num_attention_heads,
            "max_position_embeddings": max_seq_length,
            "tokenizer_type": tokenizer_type,
            "vocab_file": vocab_file,
        }

        initialize_megatron(None, megatron_args, ignore_unknown_args=True)
        init_method = init_method_normal(init_method_std)

        self.language_model, self._language_model_key = get_language_model(
            attention_mask_func=bert_attention_mask_func,
            num_tokentypes=num_tokentypes,
            add_pooler=False,
            init_method=init_method,
            scaled_init_method=scaled_init_method_normal(init_method_std, num_layers),
        )

        self.language_model.to(self._device)
        self._hidden_size = self.language_model.hidden_size

    @property
    def hidden_size(self):
        """
            Property returning hidden size.

            Returns:
                Hidden size.
        """
        return self._hidden_size

    def forward(self, input_ids, attention_mask, token_type_ids):
        extended_attention_mask = bert_extended_attention_mask(
            attention_mask, next(self.language_model.parameters()).dtype
        )
        position_ids = bert_position_ids(input_ids)

        sequence_output = self.language_model(
            input_ids, position_ids, extended_attention_mask, tokentype_ids=token_type_ids
        )
        return sequence_output

    def restore_from(self, path, local_rank=0):
        if self.placement == DeviceType.AllGpu:
            load_device = f"cuda:{local_rank}"
        else:
            load_device = self._device

        state_dict = torch.load(path, map_location=load_device)

        # to load from Megatron pretrained checkpoint
        if 'model' in state_dict:
            self.language_model.load_state_dict(state_dict['model'][self._language_model_key])
        else:
            self.load_state_dict(state_dict)



""" this code is based on PyTorch's tutorials:
https://pytorch.org/tutorials/beginner/chatbot_tutorial.html
"""
import itertools
import re
import unicodedata

import torch as t

from nemo import logging

# Default word tokens
PAD_token = 0  # Used for padding short sentences
SOS_token = 1  # Start-of-sentence token
EOS_token = 2  # End-of-sentence token


class Voc:
    def __init__(self, name):
        self.name = name
        self.trimmed = False
        self.word2index = {}
        self.word2count = {}
        self.index2word = {
            PAD_token: "PAD",
            SOS_token: "SOS",
            EOS_token: "EOS",
        }
        self.num_words = 3  # Count SOS, EOS, PAD

    def addSentence(self, sentence):
        for word in sentence.split(" "):
            self.addWord(word)

    def addWord(self, word):
        if word not in self.word2index:
            self.word2index[word] = self.num_words
            self.word2count[word] = 1
            self.index2word[self.num_words] = word
            self.num_words += 1
        else:
            self.word2count[word] += 1

    # Remove words below a certain count threshold
    def trim(self, min_count):
        if self.trimmed:
            return
        self.trimmed = True

        keep_words = []

        for k, v in self.word2count.items():
            if v >= min_count:
                keep_words.append(k)

        logging.info(
            "keep_words {} / {} = {:.4f}".format(
                len(keep_words), len(self.word2index), len(keep_words) / len(self.word2index),
            )
        )

        # Reinitialize dictionaries
        self.word2index = {}
        self.word2count = {}
        self.index2word = {
            PAD_token: "PAD",
            SOS_token: "SOS",
            EOS_token: "EOS",
        }
        self.num_words = 3  # Count default tokens

        for word in keep_words:
            self.addWord(word)


MAX_LENGTH = 10  # Maximum sentence length to consider


# Turn a Unicode string to plain ASCII, thanks to
# http://stackoverflow.com/a/518232/2809427


def unicodeToAscii(s):
    return "".join(c for c in unicodedata.normalize("NFD", s) if unicodedata.category(c) != "Mn")


# Lowercase, trim, and remove non-letter characters


def normalizeString(s):
    s = unicodeToAscii(s.lower().strip())
    s = re.sub(r"([.!?])", r" \1", s)
    s = re.sub(r"[^a-zA-Z.!?]+", r" ", s)
    s = re.sub(r"\s+", r" ", s).strip()
    return s


# Read query/response pairs and return a voc object


def readVocs(datafile, corpus_name):
    logging.info("Reading lines...")
    # Read the file and split into lines
    lines = open(datafile, encoding="utf-8").read().strip().split("\n")
    # Split every line into pairs and normalize
    pairs = [[normalizeString(s) for s in l.split("\t")] for l in lines]
    voc = Voc(corpus_name)
    return voc, pairs


# Returns True iff both sentences in a pair 'p' are under the MAX_LENGTH
# threshold


def filterPair(p):
    # Input sequences need to preserve the last word for EOS token
    return len(p[0].split(" ")) < MAX_LENGTH and len(p[1].split(" ")) < MAX_LENGTH


# Filter pairs using filterPair condition


def filterPairs(pairs):
    return [pair for pair in pairs if filterPair(pair)]


# Using the functions defined above, return a populated voc object and pairs
# list


def loadPrepareData(corpus_name, datafile):
    logging.info("Start preparing training data ...")
    voc, pairs = readVocs(datafile, corpus_name)
    logging.info("Read {!s} sentence pairs".format(len(pairs)))
    pairs = filterPairs(pairs)
    logging.info("Trimmed to {!s} sentence pairs".format(len(pairs)))
    logging.info("Counting words...")
    for pair in pairs:
        voc.addSentence(pair[0])
        voc.addSentence(pair[1])
    logging.info("Counted words:", voc.num_words)
    return voc, pairs


MIN_COUNT = 3  # Minimum word count threshold for trimming


def trimRareWords(voc, pairs, MIN_COUNT):
    # Trim words used under the MIN_COUNT from the voc
    voc.trim(MIN_COUNT)
    # Filter out pairs with trimmed words
    keep_pairs = []
    for pair in pairs:
        input_sentence = pair[0]
        output_sentence = pair[1]
        keep_input = True
        keep_output = True
        # Check input sentence
        for word in input_sentence.split(" "):
            if word not in voc.word2index:
                keep_input = False
                break
        # Check output sentence
        for word in output_sentence.split(" "):
            if word not in voc.word2index:
                keep_output = False
                break

        # Only keep pairs that do not contain trimmed word(s) in their input
        # or output sentence
        if keep_input and keep_output:
            keep_pairs.append(pair)

    # logging.info("Trimmed from {} pairs to {}, {:.4f} of total".format(len(
    # pairs), len(keep_pairs), len(keep_pairs) / len(pairs)))
    return keep_pairs


def indexesFromSentence(voc, sentence):
    return [voc.word2index[word] for word in sentence.split(" ")] + [EOS_token]


def zeroPadding(l, fillvalue=PAD_token):
    return list(itertools.zip_longest(*l, fillvalue=fillvalue))


def binaryMatrix(l):
    m = []
    for i, seq in enumerate(l):
        m.append([])
        for token in seq:
            if token == PAD_token:
                m[i].append(0)
            else:
                m[i].append(1)
    return m


# Returns padded input sequence tensor and lengths


def inputVar(l, voc):
    indexes_batch = [indexesFromSentence(voc, sentence) for sentence in l]
    lengths = t.tensor([len(indexes) for indexes in indexes_batch])
    padList = zeroPadding(indexes_batch)
    padVar = t.LongTensor(padList)
    return padVar, lengths


# Returns padded target sequence tensor, padding mask, and max target length


def outputVar(l, voc):
    indexes_batch = [indexesFromSentence(voc, sentence) for sentence in l]
    max_target_len = max([len(indexes) for indexes in indexes_batch])
    padList = zeroPadding(indexes_batch)
    mask = binaryMatrix(padList)
    mask = t.ByteTensor(mask).to(t.bool)
    padVar = t.LongTensor(padList)
    return padVar, mask, max_target_len


# Returns all items for a given batch of pairs


def batch2TrainData(voc, pair_batch):
    pair_batch.sort(key=lambda x: len(x[0].split(" ")), reverse=True)
    input_batch, output_batch = [], []
    for pair in pair_batch:
        input_batch.append(pair[0])
        output_batch.append(pair[1])
    inp, lengths = inputVar(input_batch, voc)
    output, mask, max_target_len = outputVar(output_batch, voc)
    # return inp.squeeze(-1), output.squeeze(-1)
    return inp, lengths, output, mask, max_target_len



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

# =============================================================================
# Copyright 2019 Salesforce Research.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom
# the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# =============================================================================

import torch

from nemo.backends.pytorch.nm import LossNM
from nemo.core.neural_types import LabelsType, Length, LogitsType, LossType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['MaskedLogLoss']


class MaskedLogLoss(LossNM):
    """
    Neural module which implements a cross entropy model with masking feature. It keeps just the target logit for cross entropy calculation

    Args:
        logits (float): output of the classifier
        labels (long): ground truth targets
        loss_mask (long): specifies the ones to get ignored in loss calculation


    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.

        logits: 4d tensor of logits

        labels: 3d tensor of labels

        loss_mask: specifies the words to be considered in the loss calculation

        """
        return {
            "logits": NeuralType(('B', 'T', 'D', 'D'), LogitsType()),
            "labels": NeuralType(('B', 'D', 'T'), LabelsType()),
            "length_mask": NeuralType(('B', 'D'), Length()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        # return {"loss": NeuralType(None)}
        return {"loss": NeuralType(elements_type=LossType())}

    def __init__(self):
        LossNM.__init__(self)

    def _loss_function(self, logits, labels, length_mask, eps=1e-10):
        logits_flat = logits.view(-1, logits.size(-1))
        log_probs_flat = torch.log(torch.clamp(logits_flat, min=eps))
        labels_flat = labels.view(-1, 1)
        losses_flat = -torch.gather(log_probs_flat, dim=1, index=labels_flat)
        losses = losses_flat.view(*labels.size())
        loss = self.masking(losses, length_mask)
        return loss

    @staticmethod
    def masking(losses, length_mask):
        max_len = losses.size(2)

        mask_ = torch.arange(max_len, device=length_mask.device)[None, None, :] < length_mask[:, :, None]
        mask_ = mask_.float()
        losses = losses * mask_
        loss = losses.sum() / mask_.sum()
        return loss



# Copyright (c) 2019 NVIDIA Corporation
"""
This package provides Neural Modules building blocks for building Software
2.0 projects
"""
from . import torchvision, tutorials
from .actions import PtActions
from .common import *
from .nm import DataLayerNM, LossNM, NonTrainableNM, TrainableNM



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import os
import time

import numpy as np
from matplotlib import pyplot as plt
from sklearn.metrics import classification_report, confusion_matrix, f1_score

from nemo import logging

__all__ = ['list2str', 'tensor2list', 'plot_confusion_matrix', 'tensor2numpy']


def list2str(l):
    return ' '.join([str(x) for x in l])


def tensor2list(tensor):
    return tensor.detach().cpu().tolist()


def tensor2numpy(tensor):
    return tensor.detach().cpu().numpy()


def plot_confusion_matrix(labels, preds, graph_fold, label_ids=None, normalize=False, prefix=''):
    '''
    Plot confusion matrix.
    Args:
      label_ids (dict): label to id map, for example: {'O': 0, 'LOC': 1}
      labels (list of ints): list of true labels
      preds (list of ints): list of predicted labels
      graph_fold (str): path to output folder
      normalize (bool): flag to indicate whether to normalize confusion matrix
      prefix (str): prefix for the plot name

    '''
    if label_ids is None:
        _plot_confusion_matrix(labels, preds, graph_fold)

    else:
        # remove labels from label_ids that don't appear in the dev set
        used_labels = set(labels) | set(preds)
        label_ids = {k: label_ids[k] for k, v in label_ids.items() if v in used_labels}

        ids_to_labels = {label_ids[k]: k for k in label_ids}
        classes = [ids_to_labels[id] for id in sorted(label_ids.values())]

        title = 'Confusion matrix'
        cm = confusion_matrix(labels, preds)
        if normalize:
            sums = cm.sum(axis=1)[:, np.newaxis]
            sums = np.where(sums == 0, 1, sums)
            cm = cm.astype('float') / sums
            title = 'Normalized ' + title

        fig = plt.figure()
        ax = fig.add_subplot(111)

        cax = ax.matshow(cm)
        ax.set_xticks(np.arange(-1, len(classes) + 1))
        ax.set_yticks(np.arange(-1, len(classes) + 1))
        ax.set_xticklabels([''] + classes, rotation=90)
        ax.set_yticklabels([''] + classes)
        ax.set_ylabel('True')
        ax.set_xlabel('Predicted')

        os.makedirs(graph_fold, exist_ok=True)
        fig.colorbar(cax)

        title = (prefix + ' ' + title).strip()
        plt.savefig(os.path.join(graph_fold, title + '_' + time.strftime('%Y%m%d-%H%M%S')))


def _plot_confusion_matrix(labels, preds, graph_fold):
    cm = confusion_matrix(labels, preds)
    logging.info(f'Confusion matrix:\n{cm}')
    fig = plt.figure()
    ax = fig.add_subplot(111)
    cax = ax.matshow(cm)
    plt.title('Confusion matrix of the classifier')
    fig.colorbar(cax)
    plt.xlabel('Predicted')
    plt.ylabel('True')
    os.makedirs(graph_fold, exist_ok=True)
    plt.savefig(os.path.join(graph_fold, time.strftime('%Y%m%d-%H%M%S')))


def get_classification_report(labels, preds, label_ids, output_dict=False):
    """
    Returns classification report
    """
    # remove labels from label_ids that don't appear in predictions or ground truths
    used_labels = set(labels) | set(preds)
    labels_names = [
        k + ' (label id: ' + str(v) + ')'
        for k, v in sorted(label_ids.items(), key=lambda item: item[1])
        if v in used_labels
    ]

    return classification_report(labels, preds, target_names=labels_names, digits=4, output_dict=output_dict)


def get_f1_scores(labels, preds, average_modes=['binary', 'weighted', 'macro', 'micro']):
    """
    Returns a dictionary with f1_score based on different averaging mode
    Args:
      labels (list of ints): list of true labels
      preds (list of ints): list of predicted labels
      average_modes (list): list of possible averaging types. Binary for is supported only for binary target.
    """
    f1_scores = {}
    for average in average_modes:
        f1_scores['F1 ' + average] = round(f1_score(labels, preds, average=average) * 100, 2)

    return f1_scores



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright 2019 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import os
import shutil
import tarfile
from typing import Dict
from unittest import TestCase

import numpy as np
import pytest
import torch
from ruamel.yaml import YAML

import nemo
import nemo.collections.asr as nemo_asr
from nemo.backends.pytorch.nm import DataLayerNM
from nemo.collections.nlp.nm.losses import SmoothedCrossEntropyLoss
from nemo.collections.nlp.nm.trainables.common import TokenClassifier
from nemo.core import WeightShareTransform
from nemo.core.neural_types import *
from nemo.utils import logging


@pytest.mark.usefixtures("neural_factory")
class TestWeightSharing(TestCase):
    labels = [
        "'",
        "a",
        "b",
        "c",
        "d",
        "e",
        "f",
        "g",
        "h",
        "i",
        "j",
        "k",
        "l",
        "m",
        "n",
        "o",
        "p",
        "q",
        "r",
        "s",
        "t",
        "u",
        "v",
        "w",
        "x",
        "y",
        "z",
        " ",
    ]
    manifest_filepath = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../data/asr/an4_train.json"))
    featurizer_config = {
        'window': 'hann',
        'dither': 1e-05,
        'normalize': 'per_feature',
        'frame_splicing': 1,
        'int_values': False,
        'window_stride': 0.01,
        'sample_rate': 16000,
        'features': 64,
        'n_fft': 512,
        'window_size': 0.02,
    }
    yaml = YAML(typ="safe")

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        data_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../data/"))
        logging.info("Looking up for test ASR data")
        if not os.path.exists(os.path.join(data_folder, "asr")):
            logging.info("Extracting ASR data to: {0}".format(os.path.join(data_folder, "asr")))
            tar = tarfile.open(os.path.join(data_folder, "asr.tar.gz"), "r:gz")
            tar.extractall(path=data_folder)
            tar.close()
        else:
            logging.info("ASR data found in: {0}".format(os.path.join(data_folder, "asr")))

    # @classmethod
    # def tearDownClass(cls) -> None:
    #     super().tearDownClass()
    #     data_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), "../data/"))
    #     logging.info("Looking up for test ASR data")
    #     if os.path.exists(os.path.join(data_folder, "asr")):
    #         logging.info("Removing test ASR data")
    #         shutil.rmtree(os.path.join(data_folder, "asr"))

    def __check_if_weights_are_equal(self, w1: Dict, w2: Dict):
        all_same = set(w1.keys()) == set(w2.keys())
        if not all_same:
            return False
        else:
            for key in w1.keys():
                all_same = all_same and np.array_equal(
                    w1[key][0].cpu().detach().numpy(), w2[key][0].cpu().detach().numpy(),
                )
        return all_same

    @pytest.mark.unit
    def test_TaylorNet_get_weights(self):
        tn1 = nemo.backends.pytorch.tutorials.TaylorNet(dim=4)
        tn2 = nemo.backends.pytorch.tutorials.TaylorNet(dim=4)
        # because of randomness, actual weights should be different
        self.assertFalse(self.__check_if_weights_are_equal(tn1.get_weights(), tn2.get_weights()))
        tn3 = nemo.backends.pytorch.tutorials.TaylorNet(dim=4)
        tn3.set_weights(tn1.get_weights())
        # check than weights are the same
        self.assertTrue(self.__check_if_weights_are_equal(tn1.get_weights(), tn3.get_weights()))
        # change weights on one module - another module should not change
        tn1.fc1.bias.data = torch.tensor([0.1])
        self.assertFalse(self.__check_if_weights_are_equal(tn1.get_weights(), tn3.get_weights()))

    # def test_TaylorNet_tie_weights(self):
    #     tn1 = nemo.backends.pytorch.tutorials.TaylorNet(dim=4)
    #     tn2 = nemo.backends.pytorch.tutorials.TaylorNet(dim=4)
    #     # because of randomness, actual weights should be different
    #     self.assertFalse(self.__check_if_weights_are_equal(tn1.get_weights(), tn2.get_weights()))
    #     tn2.tie_weights_with(tn1, list(tn1.get_weights().keys()))
    #     # change weights on one module - another module should change too
    #     tn2.fc1.bias.data = torch.tensor([0.1])
    #     self.assertTrue(self.__check_if_weights_are_equal(tn1.get_weights(), tn2.get_weights()))

    @pytest.mark.unit
    def test_tie_weights(self):
        class DummyDataLayer(DataLayerNM):
            def __init__(self, vocab_size):
                super().__init__()
                self.vocab_size = vocab_size

                class DummyDS(torch.utils.data.Dataset):
                    def __init__(self, vocab_size):
                        super().__init__()

                    def __getitem__(self, index):
                        model_inputs = torch.randint(high=vocab_size, size=[10])
                        model_outputs = torch.randint(high=vocab_size, size=[10])
                        return (model_inputs, model_outputs)

                    def __len__(self):
                        return 10

                self._dataset = DummyDS(vocab_size)

            @property
            def output_ports(self):
                return {
                    "model_inputs": NeuralType(('B', 'T')),
                    "model_outputs": NeuralType(('B', 'T'), LabelsType()),
                }

            def __len__(self):
                return len(self._dataset)

            @property
            def dataset(self):
                return self._dataset

            def data_iterator(self):
                pass

        voc_size = 10
        dim = 10
        embd = nemo.backends.pytorch.common.other.SequenceEmbedding(voc_size=voc_size, hidden_size=dim)
        proj = TokenClassifier(hidden_size=dim, num_classes=voc_size)
        data = DummyDataLayer(voc_size)
        loss = SmoothedCrossEntropyLoss(pad_id=0)
        embd.tie_weights_with(
            proj,
            weight_names=["embedding.weight"],
            name2name_and_transform={"embedding.weight": ("mlp.layer2.weight", WeightShareTransform.SAME)},
        )
        self.assertTrue(
            np.array_equal(embd.embedding.weight.detach().cpu().numpy(), proj.mlp.layer2.weight.detach().cpu().numpy())
        )
        _in, _out = data()
        pred = embd(input_seq=_in)
        pred = proj(hidden_states=pred)
        loss_t = loss(labels=_out, logits=pred)

        optimizer = nemo.backends.pytorch.actions.PtActions()
        optimizer.train(
            [loss_t], optimizer="sgd", optimization_params={"max_steps": 5, "lr": 0.0003},
        )

        self.assertTrue(
            np.array_equal(embd.embedding.weight.detach().cpu().numpy(), proj.mlp.layer2.weight.detach().cpu().numpy())
        )

    @pytest.mark.unit
    def test_untied_weights(self):
        class DummyDataLayer(DataLayerNM):
            def __init__(self, vocab_size):
                super().__init__()
                self.vocab_size = vocab_size

                class DummyDS(torch.utils.data.Dataset):
                    def __init__(self, vocab_size):
                        super().__init__()

                    def __getitem__(self, index):
                        model_inputs = torch.randint(high=vocab_size, size=[10])
                        model_outputs = torch.randint(high=vocab_size, size=[10])
                        return (model_inputs, model_outputs)

                    def __len__(self):
                        return 10

                self._dataset = DummyDS(vocab_size)

            @property
            def output_ports(self):
                return {
                    "model_inputs": NeuralType(('B', 'T')),
                    "model_outputs": NeuralType(('B', 'T'), LabelsType()),
                }

            def __len__(self):
                return len(self._dataset)

            @property
            def dataset(self):
                return self._dataset

            def data_iterator(self):
                pass

        voc_size = 10
        dim = 10
        embd = nemo.backends.pytorch.common.other.SequenceEmbedding(voc_size=voc_size, hidden_size=dim)
        proj = TokenClassifier(hidden_size=dim, num_classes=voc_size)
        data = DummyDataLayer(voc_size)
        loss = SmoothedCrossEntropyLoss(pad_id=0)
        # embd.tie_weights_with(
        #     proj,
        #     weight_names=["embedding.weight"],
        #     name2name_and_transform={"embedding.weight": ("mlp.layer2.weight", WeightShareTransform.SAME)},
        # )
        self.assertFalse(
            np.array_equal(embd.embedding.weight.detach().cpu().numpy(), proj.mlp.layer2.weight.detach().cpu().numpy())
        )
        _in, _out = data()
        pred = embd(input_seq=_in)
        pred = proj(hidden_states=pred)
        loss_t = loss(labels=_out, logits=pred)

        optimizer = nemo.backends.pytorch.actions.PtActions()
        optimizer.train(
            [loss_t], optimizer="sgd", optimization_params={"max_steps": 5, "lr": 0.0003},
        )

        self.assertFalse(
            np.array_equal(embd.embedding.weight.detach().cpu().numpy(), proj.mlp.layer2.weight.detach().cpu().numpy())
        )

    @pytest.mark.unit
    def test_set_weights(self):
        voc_size = 3
        dim = 2
        embd = nemo.backends.pytorch.common.SequenceEmbedding(voc_size=voc_size, hidden_size=dim)
        weights = torch.tensor(np.random.randint(0, 10, (3, 2)) * 1.0)
        name2weights = {"embedding.weight": (weights, True)}
        embd.set_weights(name2weight=name2weights)
        self.assertTrue(np.array_equal(embd.embedding.weight.detach().cpu().numpy(), weights.detach().cpu().numpy()))
        weights = torch.tensor(np.random.randint(0, 10, (3, 2)) * 1.0)
        self.assertFalse(np.array_equal(embd.embedding.weight.detach().cpu().numpy(), weights.detach().cpu().numpy()))

    @pytest.mark.unit
    @pytest.mark.run_only_on('GPU')
    def test_freeze_unfreeze_TrainableNM(self):
        path = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../data/jasper_smaller.yaml"))
        with open(path) as file:
            jasper_model_definition = self.yaml.load(file)
        dl = nemo_asr.AudioToTextDataLayer(
            # featurizer_config=self.featurizer_config,
            manifest_filepath=self.manifest_filepath,
            labels=self.labels,
            batch_size=4,
        )
        pre_process_params = {
            #'int_values': False,
            'frame_splicing': 1,
            'features': 64,
            'window_size': 0.02,
            'n_fft': 512,
            'dither': 1e-05,
            'window': 'hann',
            'sample_rate': 16000,
            'normalize': 'per_feature',
            'window_stride': 0.01,
        }
        preprocessing = nemo_asr.AudioToMelSpectrogramPreprocessor(**pre_process_params)
        jasper_encoder = nemo_asr.JasperEncoder(
            feat_in=jasper_model_definition['AudioToMelSpectrogramPreprocessor']['features'],
            **jasper_model_definition['JasperEncoder'],
        )
        jasper_decoder = nemo_asr.JasperDecoderForCTC(feat_in=1024, num_classes=len(self.labels))
        ctc_loss = nemo_asr.CTCLossNM(num_classes=len(self.labels))
        jasper_encoder.freeze()
        jasper_encoder.unfreeze(set(['encoder.4.mconv.0.conv.weight']))
        frozen_weight = jasper_encoder.encoder[1].mconv[0].conv.weight.detach().cpu().numpy()
        unfrozen_weight = jasper_encoder.encoder[4].mconv[0].conv.weight.detach().cpu().numpy()
        # jasper_decoder.unfreeze()
        # DAG
        audio_signal, a_sig_length, transcript, transcript_len = dl()
        processed_signal, p_length = preprocessing(input_signal=audio_signal, length=a_sig_length)

        encoded, encoded_len = jasper_encoder(audio_signal=processed_signal, length=p_length)
        # logging.info(jasper_encoder)
        log_probs = jasper_decoder(encoder_output=encoded)
        loss = ctc_loss(
            log_probs=log_probs, targets=transcript, input_length=encoded_len, target_length=transcript_len,
        )

        callback = nemo.core.SimpleLossLoggerCallback(
            tensors=[loss], print_func=lambda x: logging.info(f'Train Loss: {str(x[0].item())}'),
        )
        optimizer = nemo.backends.pytorch.actions.PtActions()
        optimizer.train(
            [loss], callbacks=[callback], optimizer="sgd", optimization_params={"max_steps": 5, "lr": 0.0003},
        )
        new_frozen_weight = jasper_encoder.encoder[1].mconv[0].conv.weight.data
        new_unfrozen_weight = jasper_encoder.encoder[4].mconv[0].conv.weight.data
        self.assertTrue(np.array_equal(frozen_weight, new_frozen_weight.detach().cpu().numpy()))
        self.assertFalse(np.array_equal(unfrozen_weight, new_unfrozen_weight.detach().cpu().numpy()))
